"""Vectice package.

The Vectice package is a library allowing data-scientists
to record their progress in the [Vectice app](https://docs.vectice.com/).

This package exposes essential Vectice classes and methods:

- the [autolog][vectice.autolog.autolog] method
- the [connect][vectice.Connection.connect] method
- the [Workspace][vectice.models.Workspace] class
- the [Project][vectice.models.Project] class
- the [Phase][vectice.models.Phase] class
- the [Iteration][vectice.models.Iteration] class
- the [Dataset][vectice.Dataset] class
- the [Model][vectice.Model] class
- the [Table][vectice.Table] class

    NOTE: **IMPORTANT INFORMATION**
        Vectice calculates dataframe statistics only if the dataframe contains more than a hundred rows for privacy reasons. Those statistics are calculate on a sample of a million rows and on the first 400 columns by default. All of these values can be changed on the Organization Settings page by the organization admin.

"""

from __future__ import annotations

import sys
import warnings
from importlib.util import find_spec

warnings.filterwarnings("ignore", ".*PYARROW_IGNORE_TIMEZONE")
if sys.version_info < (3, 8):
    print(
        "ImportWarning: Python 3.7 is no longer maintained by the community. Upgrade to Python 3.8 or higher to ensure full compatibility with Vectice and access to its complete features."
    )

is_h2o = find_spec("h2o") is not None
if is_h2o:
    warnings.warn(
        "h2o is installed, autolog unsaved graphs logging will be disabled. Please save h2o plots to enable the autolog graph functionality.",
        Warning,
    )

from vectice import api, autolog, models
from vectice.__version__ import __version__
from vectice.connection import Connection
from vectice.models.dataset import Dataset
from vectice.models.model import Metric, Model, Property
from vectice.models.resource import (
    BigQueryResource,
    DatabricksTableResource,
    DFResource,
    FileResource,
    GCSResource,
    NoResource,
    Resource,
    S3Resource,
    SnowflakeResource,
    SparkTableResource,
)
from vectice.models.resource.metadata import (
    DatasetSourceOrigin,
    DatasetSourceUsage,
    DatasetType,
    DBMetadata,
    File,
    FilesMetadata,
    MetadataDB,
)
from vectice.models.table import Table
from vectice.trace.vectice_trace import VecticeTrace
from vectice.utils.load_from_settings import load_from_settings
from vectice.utils.logging_utils import configure_vectice_loggers, disable_logging

trace = VecticeTrace
connect = Connection.connect

code_capture = True

"""Global code capture flag, enabled by default.

Code capture is triggered when registering a dataset or a model,
and only works when a valid Git repository is found.
Otherwise a warning is logged, telling what might be misconfigured in the repository.

Captured information include the repository name, URL, branch name, commit hash,
and whether the repository is dirty (has uncommitted changes).

Examples:
    To disable code capture globally:

    >>> import vectice
    >>> vectice.code_capture = False

    To re-enable code capture globally:

    >>> import vectice
    >>> vectice.code_capture = True
"""

code_file_capture = False

"""Global code file capture flag, disabled by default.

Controls whether code file is captured when logging asset into Vectice as part of the lineage.
When enabled (set to True), logging an asset will also add the executed file to its lineage.
Enabling this feature may increase API runtime due to file transfer. Databricks is currently not supported.


Examples:
    To enable code_file_capture globally:

    >>> import vectice
    >>> vectice.code_file_capture = True

    To re-enable code file capture globally:

    >>> import vectice
    >>> vectice.code_file_capture = False
"""

auto_extract = True
"""Global auto extraction flag, enabled by default.

Extraction is automatically performed when registering a dataset or a model with an attachment. Currently, it exclusively operates on Excel files and extracts sheets (as CSV files) and images to be referenced inside Vectice.

Examples:
    To disable auto extraction of files globally globally:

    >>> import vectice
    >>> vectice.auto_extract = False

    To re-enable auto extraction globally:

    >>> import vectice
    >>> vectice.auto_extract = True
"""

pickle_capture = True
"""Global pickle capture flag, enabled by default.

Pickle capture is triggered when registering a Vectice Model and only works when a valid predictor parameter is passed or when Autolog detects an estimator,
for example a scikit-learn regressor. The predictor is pickled and attached to the Model Version in Vectice. 

Examples:
    To disable pickle capture globally:

    >>> import vectice
    >>> vectice.pickle_capture = False

    To re-enable pickle capture globally:

    >>> import vectice
    >>> vectice.pickle_capture = True
    
    Standard API
    
    >>> from vectice import Model
    ...
    >>> my_estimator = LinearRegression()
    >>> model = Model(predictor=my_estimator)
    
    Autolog
    
    >>> from vectice import autolog
    ...
    >>> my_estimator = LinearRegression()
    >>> autolog.cell()
    
"""

configure_vectice_loggers(root_module_name=__name__)
silent = disable_logging
# Ignore all deprecation warnings by default
warnings.simplefilter("ignore", DeprecationWarning)
# But show them for vectice modules
warnings.filterwarnings("always", category=DeprecationWarning, module="vectice.*")

version = __version__

__all__ = [
    "BigQueryResource",
    "DBMetadata",
    "DFResource",
    "DatabricksTableResource",
    "Dataset",
    "DatasetSourceOrigin",
    "DatasetSourceUsage",
    "DatasetType",
    "File",
    "FileResource",
    "FilesMetadata",
    "GCSResource",
    "MetadataDB",
    "Metric",
    "Model",
    "NoResource",
    "Property",
    "Resource",
    "S3Resource",
    "SnowflakeResource",
    "SparkTableResource",
    "Table",
    "api",
    "autolog",
    "connect",
    "load_from_settings",
    "models",
    "silent",
    "version",
]



from __future__ import annotations

import logging
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Dict, Tuple

from vectice.models.resource.metadata.column_metadata import (
    BooleanStat,
    Column,
    ColumnCategoryType,
    DateStat,
    MostCommon,
    NumericalStat,
    Quantiles,
    Size,
    TextStat,
)
from vectice.models.resource.metadata.df_wrapper_resource import DataFrameWrapper
from vectice.utils.common_utils import safe_float, to_datetime_safe

if TYPE_CHECKING:
    from h2o import H2OFrame  # type: ignore[reportMissingImports]
    from pandas import DataFrame

try:
    from h2o import H2OFrame  # pyright: ignore[reportMissingImports]

    h2o_imported = True
except ImportError:
    h2o_imported = False

_logger = logging.getLogger(__name__)


class H2OFrameWrapper(DataFrameWrapper):
    def __init__(self, dataframe: H2OFrame):
        if h2o_imported is False:
            raise ImportError("H2O is not installed.")
        super().__init__(dataframe)

    def get_size(self) -> Size:
        """Get the size (rows and columns) of the H2O Frame."""
        self.rows: int = self.dataframe.nrows
        self.columns_numbers: int = self.dataframe.ncols
        return Size(rows=self.rows, columns=self.columns_numbers)

    def capture_column_schema(self) -> list[Column]:
        """Capture the schema of columns in the H2O Frame."""
        list_schema: list[Column] = []

        for col_name in self.dataframe.columns:
            col_type = self.dataframe.type(col_name)
            column_cat: ColumnCategoryType | None = None

            if col_type in ["int", "real"]:
                column_cat = ColumnCategoryType.NUMERICAL
                dtype = "numeric"
            elif col_type == "time":
                column_cat = ColumnCategoryType.DATE
                dtype = "datetime"
            elif col_type == "enum":
                levels = self.dataframe[col_name].levels()[0]
                if len(levels) == 2 and set(str(l).lower() for l in levels) <= {"true", "false", "0", "1", "yes", "no"}:
                    column_cat = ColumnCategoryType.BOOLEAN
                    dtype = "boolean"
                else:
                    column_cat = ColumnCategoryType.TEXT
                    dtype = "categorical"
            elif col_type == "string":
                column_cat = ColumnCategoryType.TEXT
                dtype = "string"
            else:
                column_cat = None
                dtype = str(col_type)

            list_schema.append(
                Column(
                    name=col_name,
                    data_type=dtype,
                    stats=None,
                    category_type=column_cat,
                )
            )

        return list_schema

    def capture_column_statistics(
        self,
        list_col_schema: list[Column],
        sample_rows_for_statistics: int,
        maximum_columns_for_statistics: int,
    ) -> list[Column]:
        """Capture statistics for columns in the H2O Frame."""
        columns: list[Column] = []
        result: Dict[str, BooleanStat | NumericalStat | TextStat | DateStat] = {}

        column_split_by_type = self.__get_list_column_split_by_type__(list_col_schema, maximum_columns_for_statistics)
        result_numerical, result_str, result_bool, result_date = self.__compute_h2o_stats__(
            column_split_by_type, sample_rows_for_statistics
        )
        result = {**result_numerical, **result_str, **result_bool, **result_date}

        for idx, col in enumerate(list_col_schema):
            if idx == maximum_columns_for_statistics:
                _logger.warning(
                    f"Statistics are only captured for the first {maximum_columns_for_statistics} columns of your dataframe."
                )

            stat = result.get(col.name) if idx < maximum_columns_for_statistics else None
            col.stats = stat
            columns.append(col)

        return columns

    def __compute_h2o_stats__(
        self, column_values: Dict[str, list], sample_rows_for_statistics: int
    ) -> Tuple[Dict[str, NumericalStat], Dict[str, TextStat], Dict[str, BooleanStat], Dict[str, DateStat]]:
        """Compute statistics for different column types."""
        result_numerical: Dict[str, NumericalStat] = {}
        result_str: Dict[str, TextStat] = {}
        result_bool: Dict[str, BooleanStat] = {}
        result_date: Dict[str, DateStat] = {}

        # Sample the frame if needed
        if sample_rows_for_statistics > 0 and sample_rows_for_statistics < self.dataframe.nrows:
            sampled_frame = self.dataframe.sample(n=sample_rows_for_statistics, seed=42)
        else:
            sampled_frame = self.dataframe

        if len(column_values["numerical"]) > 0:
            result_numerical = self.__compute_numeric_column_statistics__(sampled_frame, column_values["numerical"])

        if len(column_values["string"]) > 0:
            result_str = self.__compute_string_column_statistics__(sampled_frame, column_values["string"])

        if len(column_values["boolean"]) > 0:
            result_bool = self.__compute_boolean_column_statistics__(sampled_frame, column_values["boolean"])

        if len(column_values["date"]) > 0:
            result_date = self.__compute_date_column_statistics__(sampled_frame, column_values["date"])

        return result_numerical, result_str, result_bool, result_date

    def __get_list_column_split_by_type__(
        self, list_col_schema: list[Column], maximum_columns_for_statistics: int
    ) -> Dict[str, list]:
        """Split columns by their types."""
        column_split_by_type: Dict[str, list] = {"numerical": [], "string": [], "boolean": [], "date": []}

        for idx, col in enumerate(list_col_schema):
            if idx >= maximum_columns_for_statistics:
                break

            column_name = col.name
            if col.category_type == ColumnCategoryType.NUMERICAL:
                column_split_by_type["numerical"].append(column_name)
            elif col.category_type == ColumnCategoryType.DATE:
                column_split_by_type["date"].append(column_name)
            elif col.category_type == ColumnCategoryType.BOOLEAN:
                column_split_by_type["boolean"].append(column_name)
            elif col.category_type == ColumnCategoryType.TEXT:
                column_split_by_type["string"].append(column_name)

        return column_split_by_type

    def __compute_numeric_column_statistics__(self, frame: H2OFrame, columns: list[str]) -> Dict[str, NumericalStat]:
        """Compute statistics for numeric columns."""
        result_stats: Dict[str, NumericalStat] = {}

        for col in columns:
            col_data = frame[col]

            if not isinstance(col_data, H2OFrame):
                _logger.debug(f"Column {col} is not a valid H2OFrame. Skipping statistics computation.")
                continue
            # Basic statistics
            count = frame.nrows
            missing_count = col_data.isna().sum()
            missing_percentage = missing_count / count if count > 0 else 0.0  # pyright: ignore

            non_null_data = col_data[~col_data.isna()]
            mean, std_dev, min_val, max_val = None, None, None, None

            # Statistical measures
            if non_null_data.nrows > 0:  # pyright: ignore
                mean = non_null_data.mean()[0]  # pyright: ignore
                std_dev = non_null_data.sd()[0]  # pyright: ignore
                min_val = non_null_data.min()  # pyright: ignore
                max_val = non_null_data.max()  # pyright: ignore

            # Quantiles: returns a list of lists e.g [[0.25, 1.0], [0.5, 2.0], [0.75, 3.0]]
            quantiles = self.to_pandas(col_data.quantile([0.25, 0.5, 0.75]), header=False, use_pandas=False)
            q25 = quantiles[0][1]
            q50 = quantiles[1][1]
            q75 = quantiles[2][1]

            result_stats[col] = NumericalStat(
                mean=safe_float(mean),
                std_deviation=safe_float(std_dev),
                quantiles=Quantiles(
                    q_min=safe_float(min_val),
                    q25=safe_float(q25),
                    q50=safe_float(q50),
                    q75=safe_float(q75),
                    q_max=safe_float(max_val),
                ),
                missing=float(missing_percentage),
            )

        return result_stats

    def __compute_string_column_statistics__(self, frame: H2OFrame, columns: list[str]) -> Dict[str, TextStat]:
        """Compute statistics for string/categorical columns."""
        result_stats: Dict[str, TextStat] = {}

        for col in columns:
            col_data = frame[col]
            if not isinstance(col_data, H2OFrame):
                _logger.debug(f"Column {col} is not a valid H2OFrame. Skipping statistics computation.")
                continue

            count = frame.nrows
            missing_count = col_data.isna().sum()
            missing_percentage = missing_count / count if count > 0 else 0.0  # pyright: ignore
            unique_count = col_data.unique().nrows

            try:
                table_df = (
                    self.to_pandas(col_data.table()).sort_values("Count", ascending=False).head(3)  # pyright: ignore
                )
                most_commons = [MostCommon(str(row[col]), int(row["Count"]) / count) for _, row in table_df.iterrows()]
            except Exception as e:
                _logger.warning(f"Failed to compute most common values for column {col}: {e}")
                most_commons = []

            result_stats[col] = TextStat(
                unique=safe_float(unique_count),
                missing=float(missing_percentage),
                most_commons=most_commons,
            )

        return result_stats

    def __compute_boolean_column_statistics__(self, frame: H2OFrame, columns: list[str]) -> Dict[str, BooleanStat]:
        """Compute statistics for boolean columns."""
        result_stats: Dict[str, BooleanStat] = {}

        for col in columns:
            col_data = frame[col]
            if not isinstance(col_data, H2OFrame):
                _logger.debug(f"Column {col} is not a valid H2OFrame. Skipping statistics computation.")
                continue

            count = frame.nrows
            missing_count = col_data.isna().sum()
            missing_percentage = missing_count / count if count > 0 else 0.0  # type: ignore

            true_count = 0
            false_count = 0

            try:
                table_df = self.to_pandas(col_data.table())
                value_counts = {str(row[col]).strip().lower(): int(row["Count"]) for _, row in table_df.iterrows()}  # type: ignore

                for key, count in value_counts.items():
                    if key in {"true", "1", "yes"}:
                        true_count += count
                    elif key in {"false", "0", "no"}:
                        false_count += count

            except Exception as e:
                _logger.warning(f"Failed to compute value counts for column '{col}': {e}")
                continue

            total_non_missing = true_count + false_count
            true_pct = true_count / total_non_missing if total_non_missing else 0.0
            false_pct = false_count / total_non_missing if total_non_missing else 0.0

            result_stats[col] = BooleanStat(
                true=safe_float(true_pct),
                false=safe_float(false_pct),
                missing=safe_float(missing_percentage),
            )

        return result_stats

    def __compute_date_column_statistics__(self, frame: H2OFrame, columns: list[str]) -> Dict[str, DateStat]:
        """Compute statistics for date columns."""
        result_stats: Dict[str, DateStat] = {}

        for col in columns:
            col_data = frame[col]
            if not isinstance(col_data, H2OFrame):
                _logger.debug(f"Column {col} is not a valid H2OFrame. Skipping statistics computation.")
                continue
            count = frame.nrows
            missing_count = col_data.isna().sum()
            missing_percentage = missing_count / count if count > 0 else 0.0  # pyright: ignore

            col_data_numeric = col_data.asnumeric()
            non_null_numeric = col_data_numeric[~col_data.isna()]
            min_date = to_datetime_safe(non_null_numeric.min())  # pyright: ignore
            max_date = to_datetime_safe(non_null_numeric.max())  # pyright: ignore

            try:
                median_ts = self.to_pandas(col_data.quantile(0.5))[1][0]
                median_date = (
                    datetime.fromtimestamp(median_ts / 1000, timezone.utc) if median_ts else None  # pyright: ignore
                )
            except Exception:
                try:
                    timestamps = self.to_pandas(col_data)[col].dropna()  # pyright: ignore
                    median_ts = timestamps.median()
                    median_date = datetime.fromtimestamp(median_ts / 1000, timezone.utc)
                except Exception as e:
                    _logger.warning(f"Failed to compute median for column {col}: {e}")
                    median_date = None
            try:
                timestamps = self.to_pandas(col_data).iloc[:, 0].dropna()  # pyright: ignore
                mean_ts = timestamps.mean()
                mean_date = datetime.fromtimestamp(mean_ts / 1000, timezone.utc) if mean_ts else None
            except Exception as e:
                _logger.warning(f"Failed to compute mean for column {col}: {e}")
                mean_date = None

            result_stats[col] = DateStat(
                missing=float(missing_percentage),
                minimum=str(min_date) if min_date is not None else None,
                mean=str(mean_date) if mean_date else None,
                median=str(median_date) if median_date else None,
                maximum=str(max_date) if max_date is not None else None,
            )

        return result_stats

    @staticmethod
    def to_pandas(h2o_frame: H2OFrame, header: bool = True, use_pandas: bool = True) -> DataFrame:
        """Converts H2OFrame to Pandas with version-safe threading support."""
        import h2o  # pyright: ignore[reportMissingImports]

        version_str = h2o.__version__
        major, minor = map(int, version_str.split(".")[:2])

        kwargs = {"use_pandas": use_pandas, "header": header}
        if (major, minor) >= (3, 44):
            kwargs["use_multi_thread"] = True
        else:
            kwargs["use_single_thread"] = False

        try:
            return h2o_frame.as_data_frame(**kwargs)  # pyright: ignore
        except TypeError:
            return h2o_frame.as_data_frame(use_pandas=use_pandas)  # pyright: ignore



from __future__ import annotations

import logging
from importlib.util import find_spec
from typing import TYPE_CHECKING, Any, Callable

from vectice.autolog.asset_services.metric_service import MetricService
from vectice.autolog.asset_services.property_service import PropertyService
from vectice.autolog.asset_services.service_types.vectice_types import VecticeType
from vectice.autolog.asset_services.technique_service import TechniqueService
from vectice.autolog.model_library import ModelLibrary
from vectice.models.model import Model
from vectice.models.table import Table
from vectice.utils.common_utils import get_asset_name, temp_directory

if TYPE_CHECKING:
    from pandas import DataFrame

    from vectice.autolog.model_types import ModelTypes
    from vectice.models.table import Table

_logger = logging.getLogger(__name__)

SKLEARN_TEMP_DIR = "sklearn"
IS_PANDAS = find_spec("pandas") is not None


def identify_estimator(asset: Any) -> bool:
    """Identify if the asset is a scikit-learn or xgboost estimator."""
    estimators = {"regressor", "classifier", "clusterer"}
    # prevent seeing is_classifier and is_regressor stdout for sklearn > 1.6.0
    try:
        # sklearn > 1.6.1, xgboost does not support this yet
        estimator_type = asset.__sklearn_tags__().estimator_type
        # It's possible to have a scenario where the tags are None but implemented in XGBoost, due to the sklearn changes
        if estimator_type:
            return estimator_type in estimators

    except Exception:
        pass

    try:
        # sklearn < 1.6.1, xgboost temp fallback. Current sklearn integration is behind
        return getattr(asset, "_estimator_type", None) in estimators

    except Exception:
        pass
    return False


class AutologSklearnService(MetricService, PropertyService, TechniqueService):
    def __init__(
        self,
        key: str,
        asset: Any,
        data: dict,
        custom_metrics_data: set[str | None],
        phase_name: str,
        prefix: str | None = None,
    ):
        self._asset = asset
        self._key = key
        self._temp_dir = temp_directory(SKLEARN_TEMP_DIR)
        self._model_name = get_asset_name(self._key, phase_name, prefix)

        super().__init__(cell_data=data, custom_metrics=custom_metrics_data)

    def get_asset(self):
        """This method processes scikit-learn estimators, pipelines, and model search objects to create a standardized Model asset.
        It identifies the type of sklearn object (estimator, pipeline, or search), extracts relevant information such as parameters and metrics,
        and generates appropriate attachments based on the model type. For search objects, it extracts search results and hyperparameter space tables.
        For pipelines, it generates HTML and JSON representations.
        For regular estimators, it extracts model parameters and metrics.

        Note:
        Get the model asset if it is a valid sklearn or xgboost model. xgboost relies on BaseEstimator.
        lightgbm has Booster and sklearn API which uses BaseEstimator.
        """
        from sklearn.pipeline import Pipeline

        temp_files = []
        is_estimator = identify_estimator(self._asset)
        is_pipeline = isinstance(self._asset, Pipeline)
        is_search = is_estimator and str(self._asset.__class__.__module__) == "sklearn.model_selection._search"

        if not (is_estimator or is_pipeline or is_search):
            return None

        if is_search:
            library = ModelLibrary.SKLEARN_SEARCH
            tables = self._get_sklearn_search_results_and_space_tables(self._asset, self._model_name)
            temp_files.extend(tables)

        elif is_pipeline:
            library = ModelLibrary.SKLEARN_PIPELINE
            html_and_json_paths = self._get_sklearn_pipeline_html_and_json(self._asset, self._model_name)
            temp_files.extend(html_and_json_paths)

        else:
            library = ModelLibrary.SKLEARN

        try:
            # TODO fix regex picking up classes
            # Ignore Initialized variables e.g LogisticRegression Class
            self._asset.get_params()  # pyright: ignore[reportGeneralTypeIssues]
            _, params = self._get_sklearn_or_xgboost_or_lgbm_info(self._asset)
            model = Model(
                library=library.value,
                technique=self._get_model_technique(self._asset, ModelLibrary.SKLEARN),
                metrics=self._get_model_metrics(self._cell_data),
                properties=params,
                name=self._model_name,
                predictor=self._asset,
                attachments=temp_files,
            )
            return {
                "variable": self._key,
                "model": model,
                "asset_type": VecticeType.MODEL,
            }
        except Exception as e:
            _logger.debug(f"Failed to get asset from {self._asset.__class__.__name__}: {e!s}.")

    def _get_sklearn_pipeline_html_and_json(self, pipeline: ModelTypes, model_name: str) -> list[str]:
        """Generate JSON and HTML representations of sklearn pipeline."""
        from sklearn.utils import estimator_html_repr

        from vectice.utils.sklearn_pipe_utils import pipeline_to_json

        file_generators = [("json", "json", pipeline_to_json), ("html", "utf-8", estimator_html_repr)]

        return [
            file_path
            for extension, encoding, generator in file_generators
            if (file_path := self._create_pipeline_file(pipeline, model_name, extension, encoding, generator))
        ]

    def _create_pipeline_file(
        self,
        pipeline: ModelTypes,
        model_name: str,
        extension: str,
        encoding: str,
        generator_func: Callable[[ModelTypes], str | None],
    ) -> str | None:
        """Create a pipeline representation file."""
        try:
            content = generator_func(pipeline)
            if not content:
                return None

            file_path = self._temp_dir / f"{model_name}_pipeline.{extension}"
            with open(file_path, "w", encoding=encoding) as file:
                file.write(content)

            return file_path.as_posix()

        except Exception as e:
            _logger.debug(f"Failed to create pipeline {extension.upper()} for {model_name}: {e}")
            return None

    def _get_sklearn_search_results_and_space_tables(self, model: ModelTypes, model_name: str) -> list[Table]:
        if not IS_PANDAS:
            _logger.debug("Pandas is not available, cannot process sklearn search results and space tables.")
            return []

        import pandas as pd

        tables = []
        results_df = pd.DataFrame(model.cv_results_)  # pyright: ignore

        try:
            tables.append(self._get_sklearn_search_space(model_name, results_df))
        except Exception as e:
            _logger.debug(f"Failed to get sklearn search space table for {model_name}: {e}")
        try:
            tables.append(self._get_sklearn_search_results(model_name, results_df))
        except Exception as e:
            _logger.debug(f"Failed to get sklearn search results table for {model_name}: {e}")
        return tables

    def _get_sklearn_search_results(self, model_name: str, results_df: DataFrame) -> Table:
        sorted_df = results_df.sort_values(by="rank_test_score")
        top_scores_df = sorted_df.head(5)
        return Table(top_scores_df, name=f"{model_name}_search_results")

    def _get_sklearn_search_space(self, model_name: str, results_df: DataFrame) -> Table:
        import pandas as pd

        param_columns = [col for col in results_df.columns if "param" in col and "params" not in col]
        data_dict = {}
        for param in param_columns:
            try:
                data_dict[f"{param} (min,max)"] = [[results_df[param].min(), results_df[param].max()]]
            except Exception:
                data_dict[f"{param} (uniques)"] = [results_df[param].unique()]

        params_df = pd.DataFrame(dict([(k, pd.Series(v)) for k, v in data_dict.items()]))
        # Transpose the DataFrame
        df_transposed = params_df.transpose()
        # Reset the index to move row headers into a column
        df_transposed_reset = df_transposed.reset_index()
        # Rename the index column for clarity
        df_transposed_reset = df_transposed_reset.rename(columns={"index": "Parameters", 0: "Values"})
        return Table(df_transposed_reset, name=f"{model_name}_search_space")



from __future__ import annotations

from typing import Any

from vectice.autolog.asset_services.service_types.giskard_types import GiskardType
from vectice.utils.common_utils import temp_directory

GISKARD_TEMP_DIR = "giskard"


class GiskardTestSuiteResultService:
    def __init__(self, key: str, asset: Any):
        self._asset = asset
        self._key = key
        self._temp_dir = temp_directory(GISKARD_TEMP_DIR)

    def get_asset(self):
        asset = {
            "variable": self._key,
            "suite_name": self._asset.suite.name,
            "attachment": None,
            "asset_type": GiskardType.TEST_SUITE_RESULT,
        }
        html_content = self._asset._repr_html_()

        filename_path = self._temp_dir / f"{self._key}_test_suite.html"
        filename = filename_path.as_posix()

        with open(filename, "w") as file:
            file.write(html_content)

        asset["attachment"] = filename
        return asset



from __future__ import annotations

import logging
import math
from abc import abstractmethod
from typing import Generic, TypeVar

from pandas import CategoricalDtype, Series, api

from vectice.models.resource.metadata.column_metadata import (
    BooleanStat,
    Column,
    ColumnCategoryType,
    DateStat,
    MostCommon,
    NumericalStat,
    Quantiles,
    Size,
    TextStat,
)
from vectice.models.resource.metadata.dataframe_config import (
    DataFramePandasType,
)
from vectice.models.resource.metadata.df_wrapper_resource import DataFrameWrapper

_logger = logging.getLogger(__name__)

PT = TypeVar("PT", bound=DataFramePandasType)


class DataFrameDefaultPandasWrapper(DataFrameWrapper[PT], Generic[PT]):
    rows: int
    columns_numbers: int

    def __init__(self, dataframe: PT):
        super().__init__(dataframe)
        self.rows, self.columns_numbers = self.dataframe.shape

    @abstractmethod
    def __compute_date_column_statistics__(self, series: Series) -> DateStat:
        pass

    def get_size(self) -> Size:
        return Size(rows=int(self.rows), columns=int(self.columns_numbers))

    def is_date_series(self, column_type: str) -> bool:
        if api.types.is_datetime64_any_dtype(column_type):
            return True

        try:
            # Temporary fixing issue -> TypeError: data type 'dbdate' not understood [EN-2534]
            if column_type == "dbdate":
                return True
        except TypeError:
            pass

        return False

    def capture_column_schema(self) -> list[Column]:
        column_cat: ColumnCategoryType | None = None
        list_schema: list[Column] = []

        for column, column_type in self.dataframe.dtypes.items():
            column = str(column)
            if api.types.is_bool_dtype(column_type):
                column_cat = ColumnCategoryType.BOOLEAN
                dtype = str(column_type)
            elif api.types.is_numeric_dtype(column_type):
                column_cat = ColumnCategoryType.NUMERICAL
                dtype = str(column_type)
            elif self.is_date_series(column_type):
                column_cat = ColumnCategoryType.DATE
                dtype = str(column_type)
            # Assign categorical columns as strings
            elif api.types.is_string_dtype(column_type) or isinstance(column_type, CategoricalDtype):
                column_cat = ColumnCategoryType.TEXT
                dtype = str(column_type)
            else:
                column_cat = None
                dtype = str(column_type)

            list_schema.append(
                Column(
                    name=column,
                    data_type=dtype if dtype != "object" else "string",
                    stats=None,
                    category_type=column_cat,
                )
            )
        return list_schema

    def capture_column_statistics(
        self,
        list_col_schema: list[Column],
        sample_rows_for_statistics: int,
        maximum_columns_for_statistics: int,
    ) -> list[Column]:
        columns: list[Column] = []
        stat: TextStat | BooleanStat | NumericalStat | DateStat | None = None

        from vectice.models.resource.metadata.dataframe_config import MAX_COLUMNS_CAPTURE_STATS

        if sample_rows_for_statistics < len(self.dataframe):
            sampled_dataframe = self.dataframe.sample(n=sample_rows_for_statistics, random_state=42)
        else:
            sampled_dataframe = self.dataframe
        failed_stats_cols: list[str] = []
        for idx, col in enumerate(list_col_schema):
            if idx == maximum_columns_for_statistics:
                _logger.warning(
                    f"Statistics are only captured for the first {maximum_columns_for_statistics or MAX_COLUMNS_CAPTURE_STATS} columns of your dataframe."
                )
            column = sampled_dataframe[col.name]
            col_type = col.category_type
            if col_type is not None and (idx < maximum_columns_for_statistics):
                try:
                    stat = self.__capture_pandas_stats__(column, col_type)
                except Exception as e:
                    stat = None
                    _logger.debug(f"Failed to capture statistics for column '{col.name}'. Reason: {e}")
                    failed_stats_cols.append(col.name)
            else:
                stat = None
            col.stats = stat

            columns.append(col)

        if failed_stats_cols:
            _logger.warning(
                f"Statistics could not be captured for the following columns: {', '.join(failed_stats_cols)}. "
                "Please ensure these columns do not contain mixed data types."
            )

        return columns

    def __capture_pandas_stats__(
        self, series: Series, col_type: ColumnCategoryType
    ) -> TextStat | BooleanStat | NumericalStat | DateStat | None:
        if col_type.value == "BOOLEAN":
            return self.__compute_boolean_column_statistics__(series)
        elif col_type.value == "NUMERICAL":
            return self.__compute_numeric_column_statistics__(series)
        elif col_type.value == "DATE":
            return self.__compute_date_column_statistics__(series)
        elif col_type.value == "TEXT":
            return self.__compute_string_column_statistics__(series)
        return None

    def __compute_boolean_column_statistics__(self, series: Series) -> BooleanStat:
        """Parse a dataframe series and return statistics about it.

        The computed statistics are:
        - The percentage of True
        - The percentage of False
        - The count missing value in %
        Parameters:
            series: The pandas series to get information from.

        Returns:
            A BooleanStat object containing the above statistics.
        """
        value_counts = series.value_counts(dropna=False)
        value_counts_percentage = value_counts / value_counts.sum()
        series_missing = series.isnull().sum()
        missing = series_missing / len(series)
        return BooleanStat(
            true=float(value_counts_percentage.get(True, 0)),
            false=float(value_counts_percentage.get(False, 0)),
            missing=float(missing) if not (math.isnan(missing) or math.isinf(missing)) else None,
        )

    def _replace_nan_values(self, *args):  # pyright: ignore[reportMissingParameterType]
        """Handle NaN column statistic values."""
        return tuple(None if math.isnan(x) or math.isinf(x) else float(x) for x in args)

    def _replace_str_nan_values(self, *args):  # pyright: ignore[reportMissingParameterType]
        """Handle NaN column statistic values."""
        return tuple(None if math.isnan(x) or math.isinf(x) else str(x) for x in args)

    def __compute_numeric_column_statistics__(self, series: Series) -> NumericalStat:
        """Parse a dataframe series and return statistics about it.

        The computed statistics are:
        - the mean
        - the standard deviation
        - the min value
        - the 25% percentiles
        - the 50% percentiles
        - the 75% percentiles
        - the max value
        - the count missing value in %
        Parameters:
            series: The pandas series to get information from.

        Returns:
            A NumericalStat object containing the above statistics.
        """
        mean = series.mean()
        std = series.std()
        min_q = series.min()
        q25 = series.quantile(0.25)
        q50 = series.quantile(0.5)
        q75 = series.quantile(0.75)
        max_q = series.max()
        missing = float(series.isnull().sum()) / float(len(series))
        mean, std, min_q, q25, q50, q75, max_q = self._replace_nan_values(mean, std, min_q, q25, q50, q75, max_q)
        return NumericalStat(
            mean=mean,
            std_deviation=std,
            quantiles=Quantiles(q_min=min_q, q25=q25, q50=q50, q75=q75, q_max=max_q),
            missing=missing,
        )

    def __compute_string_column_statistics__(self, series: Series) -> TextStat:
        """Parse a dataframe series and return statistics about it.

        The computed statistics are:
        - the unique number of value
        - the top 3 most common values with their percentages
        - the count missing value in %
        Parameters:
            series: The pandas series to get information from.

        Returns:
            A TextStat object containing the above statistics.
        """
        missing = series.isnull().sum() / len(series)
        unique = len(series.unique())
        value_counts = series.value_counts() / series.value_counts(dropna=False).sum()

        size = 3 if unique >= 3 else unique
        value_counts_largest: Series[float] = value_counts.nlargest(size)
        missing, unique = self._replace_nan_values(missing, unique)
        return TextStat(
            unique=unique,
            missing=missing,
            most_commons=[
                MostCommon(str(i), float(value_counts_largest[i])) for i in value_counts_largest.index.values
            ],
        )



from __future__ import annotations

import ast
import logging
import re
import uuid
from collections import OrderedDict
from importlib.util import find_spec
from typing import TYPE_CHECKING, Any, Callable, TypeVar

import matplotlib.pyplot as plt
from typing_extensions import TypedDict

from vectice.api.http_error_handlers import OrganizeError, VecticeException
from vectice.autolog.asset_services.service_types import GiskardType, ModevaType, VecticeType
from vectice.autolog.autolog_asset_factory import AssetFactory
from vectice.models.dataset import Dataset
from vectice.models.model import Model
from vectice.models.representation.dataset_representation import DatasetRepresentation
from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
from vectice.models.representation.model_representation import ModelRepresentation
from vectice.models.representation.model_version_representation import ModelVersionRepresentation
from vectice.models.table import Table
from vectice.models.validation import ValidationModel
from vectice.utils.code_parser import VariableVisitor, parse_comments, preprocess_code
from vectice.utils.common_utils import ensure_correct_project_id_from_representation_objs
from vectice.utils.last_assets import _get_asset_parent_name  # pyright: ignore [reportPrivateUsage]

if TYPE_CHECKING:
    from modeva.utils.results import ValidationResult  # type: ignore[reportMissingImports]
    from pandas import DataFrame
    from pyspark.sql import DataFrame as SparkDF
    from sklearn.pipeline import Pipeline

    # Vectice Object types
    from vectice.autolog.asset_services import TVecticeObjects
    from vectice.models import Phase
    from vectice.models.resource.metadata.db_metadata import TableType

    # container for metric tables and future floating assets
    TMetric = TypedDict(
        "TMetric",
        {
            "variable": str,
            "tables": list[Table | None],
            "asset_type": VecticeType,
        },
    )

    TModel = TypedDict(
        "TModel",
        {
            "variable": str,
            "model": Model,
            "summary": str,
            "asset_type": VecticeType,
        },
    )

    DataframeTypes = TypeVar("DataframeTypes", SparkDF, DataFrame)
    TDataset = TypedDict(
        "TDataset",
        {
            "variable": str,
            "dataset": Dataset,
            "asset_type": VecticeType,
        },
    )
    TValidationResult = TypedDict(
        "TValidationResult",
        {"variable": str, "table": DataframeTypes, "plot": TableType, "asset": ValidationResult},
    )
    TFactSheet = TypedDict(
        "TFactSheet",
        {"dataset": DataframeTypes, "models": list, "validation_results": list[TValidationResult]},
    )

    TWidgetOutput = TypedDict(
        "TWidgetOutput",
        {
            "variable": str,
            "output": Any,
            "asset_type": VecticeType,
        },
    )


try:
    from IPython.core.getipython import get_ipython
    from IPython.core.interactiveshell import InteractiveShell
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "To use autolog, please install extra dependencies from vectice package using '%pip install vectice[autolog]'"
    ) from None

is_plotly = False
is_matplotlib = False
if find_spec("plotly") is not None:
    is_plotly = True

if find_spec("matplotlib") is not None:
    is_matplotlib = True

_logger = logging.getLogger(__name__)

NOTEBOOK_CELLS = {}
GRAPHS = {}
IPYTHON_SESSION = None


def validate_ipython_session(ipython: Any) -> bool:
    """Validate IPython session. As opening and closing the browser will create a new header and session. The cellIds become untrackable."""
    global IPYTHON_SESSION
    global GRAPHS

    ipython_session_id = ipython.get_parent()["header"]["session"]  # type: ignore
    if IPYTHON_SESSION and IPYTHON_SESSION != ipython_session_id:
        GRAPHS = {}  # pyright: ignore[reportConstantRedefinition]
        IPYTHON_SESSION = ipython_session_id  # pyright: ignore[reportConstantRedefinition]
        return False

    IPYTHON_SESSION = ipython_session_id  # pyright: ignore[reportConstantRedefinition]
    return True


# TODO - migrate to our vectice id for all notebooks
def _inject_vectice_cell_id(ipython: Any) -> str:
    """Injects a cell id into the IPython metadata."""
    ipython_metadata = ipython.get_parent()  # type: ignore
    if ipython_metadata.get("metadata", {}).get("vecticeId", False):  # type: ignore
        return ipython_metadata["metadata"]["vecticeId"]  # type: ignore
    vectice_id = str(uuid.uuid4())
    # Inject the vecticeId into the metadata
    ipython.get_parent()["metadata"]["vecticeId"] = vectice_id  # type: ignore
    return vectice_id


def _get_cell_id(ipython: Any):
    try:
        ipython_metadata = ipython.get_parent()  # type: ignore
        cell_metadata = ipython_metadata.get("metadata", {})  # type: ignore
        _logger.debug(f"Cell Metadata {cell_metadata}")
        if cell_metadata.get("colab") and "cell_id" in cell_metadata.get("colab"):  # type: ignore
            cell_id = cell_metadata["colab"]["cell_id"]
        elif cell_metadata.get("commandId"):  # type: ignore
            cell_id = cell_metadata["commandId"]
        elif ipython_metadata.get("cellId", False):  # type: ignore
            cell_id = ipython_metadata["cellId"]
        else:
            # if no cellId is found, inject a new one. Persists in Ipython metadata
            _logger.debug("No cellId found in metadata, injecting a new VecticeId.")
            cell_id = _inject_vectice_cell_id(ipython)  # type: ignore
        return cell_id
    except Exception as e:
        _logger.debug(f"Unable to retrieve CellId.\n{e}")
    return None


def start_listen() -> Any:
    ipython = get_ipython()  # type: ignore

    def _notebook_cell_content_update():
        global NOTEBOOK_CELLS

        cell_id = _get_cell_id(ipython)
        try:
            cell_content = ipython.get_parent()["content"]["code"]  # type: ignore
            NOTEBOOK_CELLS[cell_id] = cell_content
        except Exception as e:
            _logger.debug(f"Notebook cell content failed to update.\n{e}")

    def _setup_cell():
        # if a cell is run or re-run, clear the dictionary
        cell_id = _get_cell_id(ipython)
        if cell_id:
            GRAPHS[cell_id] = {
                "saved": [],
                "displayed": [],
            }

    def _load_ipython_extension():
        is_inline = True
        try:
            from matplotlib_inline.backend_inline import flush_figures
        except ImportError:
            is_inline = False

        from vectice.utils.plot_tracker import PlotTracker

        # Connect the plot tracker hook to IPython's display hook
        # NB we monkey patch plotly if package is found, see in plot_tracker.py
        plot_tracker = PlotTracker(is_inline)
        # Apply the patch for savefig
        plt.savefig = plot_tracker.patched_savefig

        # unregister the matplotlib inline event
        # this is done so the graphs aren't flushed before we can capture them
        is_flush_figure_registered = get_event_callback(ipython, "post_execute", "flush_figures")
        if is_flush_figure_registered:
            ipython.events.unregister("post_execute", flush_figures)  # type: ignore

        """Load the extension in IPython."""
        ipython.events.register("pre_execute", _setup_cell)  # type: ignore
        ipython.events.register("post_execute", _notebook_cell_content_update)  # type: ignore
        ipython.events.register("post_execute", plot_tracker.get_all_figures)  # type: ignore
        # register the matplotlib event so we don't break the display output
        if is_flush_figure_registered:
            ipython.events.register("post_execute", flush_figures)  # type: ignore

    pre_execute_callback = get_event_callback(ipython, "pre_execute", "_setup_cell")
    post_execute_callback = get_event_callback(ipython, "post_execute", "_notebook_cell_content_update")
    post_run_cell_callback = get_event_callback(ipython, "post_execute", "get_all_figures")
    if ipython and not (pre_execute_callback and post_execute_callback and post_run_cell_callback):  # type: ignore
        _load_ipython_extension()

    return ipython


def get_event_callback(ipython: InteractiveShell | None, event_name: str, callback_name: str) -> Callable | None:
    """Helper function to find a registered callback by name."""
    if ipython:
        event_callbacks = [ev for ev in ipython.events.callbacks[event_name] if callback_name in str(ev)]
        return event_callbacks[0] if event_callbacks else None
    return None


####### Autolog logic
class Autolog:
    def __init__(
        self,
        phase: Phase | None,
        ipy: InteractiveShell,
        is_notebook: bool,
        create: bool = True,
        note: str | None = None,
        capture_schema_only: bool = True,
        capture_comments: bool = True,
        prefix: str | None = None,
        capture_widget_graphs: bool = False,
        organize_with_ai: bool = False,
        generate_report: bool = False,
    ):
        from vectice.services.phase_service import PhaseService

        if phase is None:
            raise ValueError("Login")

        if create is True:
            iteration = phase.create_iteration()
        else:
            iteration = PhaseService(
                phase._client  # pyright: ignore[reportPrivateUsage]
            ).get_active_iteration_or_create(phase)

        self._iteration = iteration
        self._iteration_service = iteration._service  # pyright: ignore[reportPrivateUsage]
        self._capture_schema_only = capture_schema_only
        self._ip = ipy
        self._local_vars = self._ip.user_global_ns
        self._metric_functions = self._get_metric_functions()
        self._capture_comments = capture_comments
        self._prefix = prefix
        self._is_notebook = is_notebook
        self._cell_content = self._get_notebook_cell_content() if is_notebook is True else self._get_cell_content()
        self._vectice_data = self._get_variable_matches(self._local_vars)
        self._failed_assets = []

        # Get back objects to log
        self._assets = self._get_assets()
        graphs = self._get_graphs(is_notebook, capture_widget_graphs)

        if note:
            self._iteration_service.log_comment(note)
            _logger.info(f"Note logged in iteration {self._iteration.name!r}.")

        # Log objects

        self._log_assets(self._assets)

        if graphs:
            for graph in graphs:
                try:
                    self._iteration_service.log_image_or_file(graph)
                    graph = graph if isinstance(graph, str) else graph.filename
                    _logger.info(f"Graph {graph!r} logged in iteration {self._iteration.name!r}.")
                except Exception as e:
                    self._failed_assets.append({"reason": e, "asset": graph, "type": "Graph"})

        if len(self._failed_assets):
            _logger.warning("The following assets failed to log:")
        for failed_asset in self._failed_assets:
            _logger.warning(f"{failed_asset['type']} {failed_asset['asset']!r}, reason: {failed_asset['reason']}")
        # clear metrics - prevents non capture
        self._clear_metric_assets()

        cells = [{"cellId": key, "content": value} for key, value in NOTEBOOK_CELLS.items()]
        self._iteration_service.save_autolog_assets(
            list(filter(lambda cell: cell["cellId"] is not None, cells)),  # type: ignore
            prefix,
        )
        if organize_with_ai:
            try:
                if generate_report:
                    self._iteration_service.organize_with_ai_and_report()
                else:
                    self._iteration_service.organize_with_ai()
            except OrganizeError as e:
                _logger.warning(
                    """
                    Unable to organize this iteration with AI. This feature requires code cells, which may be missing for one of the following reasons:
                        1. Code upload is not enabled for your organization.
                        2. This iteration was not created using Autolog.
                    If you'd like to use 'Organize with Ask AI', please contact your organization admin to enable code upload and ensure the iteration is generated with Autolog.
                """
                )

        # notebook_path = get_notebook_path()
        # if save_notebook and notebook_path:
        #     self._iteration._log_image_or_file(notebook_path, section="Notebook")  # type: ignore[reportPrivateUsage]

    def _get_metric_functions(self) -> set[str | None]:
        safe_keys = set()
        SparkDataFrame = None

        try:
            from pyspark.sql import DataFrame as PySparkDataFrame

            SparkDataFrame = PySparkDataFrame
        except ImportError:
            pass  # Spark not available

        for name, func in self._local_vars.items():
            try:
                if SparkDataFrame and isinstance(func, SparkDataFrame):
                    continue
                if getattr(func, "custom_metric_function", False):
                    safe_keys.add(name)
            except Exception as e:
                _logger.debug(f"Skipping variable '{name}' during metric function check: {e}")
                continue

        return safe_keys

    def _get_variable_matches(self, local_vars: dict[str, Any]) -> list[dict[Any, Any]]:
        vectice_data = []
        all_cell_vars = set()
        all_vectice_calls = set()
        for cell_not_processed in self._cell_content:
            cell = preprocess_code(cell_not_processed)
            variable_comments = parse_comments(cell_not_processed) if self._capture_comments else []
            vectice_match = {}
            all_vars: OrderedDict[str, Any] = OrderedDict()

            vectice_match["cell"] = cell
            vectice_match["comments"] = variable_comments

            # Parse the cell content using the VariableVisitor
            try:
                tree = ast.parse(cell)
            except SyntaxError as err:
                raise SyntaxError(
                    "Autolog is unable to parse the code. Make sure all non-Python syntax, such as bash commands, are properly indented and preceded by '%' or '!'. If the error persist, please contact your sales representative."
                ) from err

            visitor = VariableVisitor(custom_metrics=self._metric_functions)
            visitor.visit(tree)
            # Keep set of all vectice call vars
            all_vectice_calls = all_vectice_calls.union(visitor.vectice_call_vars)
            # Update all_vars with variable names and values in the order they appear
            for var in visitor.variables:
                if var in local_vars and var not in all_vars:
                    all_vars[var] = local_vars[var]
                # check if the var exists in a previous cell for autolog.notebook before the vectice object uses the var
                if self._is_notebook:
                    all_vectice_calls = {vect_var for vect_var in all_vectice_calls if vect_var not in all_cell_vars}
                # Keep track of all vars
                all_cell_vars.add(var)

            vectice_match["variables"] = all_vars
            vectice_data.append(vectice_match)
        clean_vectice_data = []
        for data in vectice_data:
            for var in all_vectice_calls:
                # the order of cells misses vars
                if var in data["variables"]:
                    del data["variables"][var]
            # check that there are vars for the cell and then append or if there are comments
            if len(data["variables"]) >= 1 or len(data["comments"]) >= 1:
                clean_vectice_data.append(data)
        return clean_vectice_data

    def _get_notebook_cell_content(self) -> list[Any]:
        try:
            cell_id = _get_cell_id(self._ip)
            cell_content = self._ip.get_parent()["content"]["code"]  # type: ignore
            NOTEBOOK_CELLS[cell_id] = cell_content
        except Exception as e:
            _logger.debug(f"Failed to get notebook cell content.\n{e}")
        return list(NOTEBOOK_CELLS.values())

    def _get_cell_content(self) -> list[Any]:
        """Used by autolog cell to get the content of the cell. This is used to parse for variables."""
        cell_content = self._ip.get_parent()["content"]["code"]  # pyright: ignore[reportAttributeAccessIssue]
        if cell_content is None:
            raise ValueError("Failed to get cell content.")
        return [cell_content]

    def _get_asset_comments(self, comments: dict, key: str) -> tuple[str | None, str | None]:
        try:
            comment_before = comments["comment_matches_before"][key]["comment"]
        except KeyError:
            comment_before = None
        try:
            comment_after = comments["comment_matches_after"][key]["comment"]
        except KeyError:
            comment_after = None
        return comment_before, comment_after

    def _get_assets(self) -> list[TModel | TDataset | TVecticeObjects | TWidgetOutput]:
        """Collects assets by processing vectice data and deduplicates them."""
        assets = []
        phase_name = self._iteration.phase.id
        processed_keys = set()

        for data in self._vectice_data:
            # Process variables
            for key, asset in data["variables"].items():
                if not key.startswith("_") and key not in processed_keys:  # Skip cell inputs/outputs
                    self._process_variable(key, asset, data, phase_name, assets)
                    processed_keys.add(key)

            # Process metrics
            self._process_metric(data, assets)

        return self._deduplicate_assets(assets)

    def _process_variable(self, key: str, asset: Any, data: dict, phase_name: str, assets: list[dict[str, Any]]):
        """Processes a single variable and adds its asset to the list if valid."""
        try:
            asset_service = AssetFactory.get_asset_service(
                key,
                asset,
                data,
                phase_name,
                self._metric_functions,
                self._cell_content,
                self._capture_schema_only,
                self._prefix,
            )
            asset_information = asset_service.get_asset()
            if asset_information:
                assets.append(asset_information)
        except VecticeException:
            pass

    def _process_metric(self, data: dict, assets: list[dict[str, Any]]):
        from vectice.autolog.asset_services.metric_service import MetricService

        try:
            metric_asset = MetricService(data, self._metric_functions).get_assets()
            if metric_asset:
                assets.append(metric_asset)
        except VecticeException:
            pass

    def _deduplicate_assets(
        self, assets: list[dict[str, Any]]
    ) -> list[TModel | TDataset | TVecticeObjects | TWidgetOutput]:
        unique_dict = {}
        result = []

        for item in assets:
            variable = item["variable"]

            if variable not in unique_dict.keys():
                # If variable is not seen before, add it to the dictionary
                unique_dict[variable] = item
                result.append(item)
            else:
                existing_item = unique_dict[variable]
                if "vectice_object" in item:
                    existing_item["vectice_object"] = item["vectice_object"]
                elif "dataframe" in item:
                    existing_item["type"] = item["type"]
                elif "library" in item:
                    existing_item["library"] = item["library"]

                    # Merge dictionaries for the "metrics" attribute
                    existing_item_metrics = existing_item["metrics"]
                    item_metrics = item["metrics"]

                    for key, value in item_metrics.items():
                        existing_item_metrics[key] = value
                elif "validation_results" in item:
                    existing_item["dataset"] = item["dataset"]
                    existing_item["models"] = item["models"]
                    existing_item["validation_results"] = item["validation_results"]
                elif "table" in item:
                    existing_item["table"] = item["table"]
                    existing_item["plot"] = item["plot"]
                elif "tables" in item:
                    existing_item["tables"] = item["tables"]
        return result

    def _clear_metric_assets(self):
        # clear metric assets after autolog
        from vectice.autolog.asset_services.metric_service import captured_metrics

        captured_metrics.clear()

    def _get_pipeline_steps(self, pipeline: Pipeline) -> dict:
        sklearn_pipeline = {}
        try:
            for step in pipeline.steps:
                step_name, step_obj = step
                if hasattr(step_obj, "named_transformers_"):
                    for k, v in step_obj.named_transformers_.items():
                        sklearn_pipeline[f"pipeline_{step_name}_{k}_steps"] = list(v.named_steps.keys())
            return sklearn_pipeline
        except Exception:
            return sklearn_pipeline

    def _get_pipeline_info(self, pipeline: Pipeline) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        # if pipeline with classifier
        try:
            from sklearn.base import is_classifier, is_regressor

            model = pipeline.steps[-1][-1]
            model_name = pipeline.steps[-1][0]
            if is_regressor(model) or is_classifier(model):
                model_params = {
                    f"{model_name}_{key}": value
                    for key, value in model.get_params().items()
                    if value is not None and bool(str(value))
                }
                model_params.update(self._get_pipeline_steps(pipeline))
                return "sklearn-pipeline", model_params
        except Exception:
            pass

        try:
            pipeline_params = {
                str(key): value
                for key, value in pipeline.get_params().items()
                if value is not None and bool(str(value))
            }
            return "sklearn-pipeline", pipeline_params
        except Exception:
            return None, None

    def _get_all_cells_comments(self) -> list[dict]:
        # Get all comments from all cells
        all_comments_and_variables = []
        for data in self._vectice_data:
            if data["comments"]:
                for comments in data["comments"]:
                    all_comments_and_variables.append(comments)
        return all_comments_and_variables

    def _get_comments_before(
        self,
        all_comments_and_variables: list[dict],
        asset: TModel | TDataset | TVecticeObjects | TWidgetOutput,
        comments_logged: list,
    ):
        comments_to_log = []
        for data in all_comments_and_variables:
            # If we find the variable we log all the comments found
            if data["variable"] == asset["variable"]:
                return comments_to_log
            # We only want comments remaining
            if data["comment"] not in comments_logged:
                comments_to_log.append(data["comment"])
        # return nothing if we don't see the asset
        return []

    def _log_comments_before_asset(
        self,
        all_comments_and_variables: list[dict],
        asset: TModel | TDataset | TVecticeObjects | TWidgetOutput,
        comments_logged: list,
    ) -> list:
        # Get comments to log before each asset
        comments_before = self._get_comments_before(all_comments_and_variables, asset, comments_logged)
        # Log comments before each asset
        if comments_before:
            for comment in comments_before:
                if comment:
                    self._iteration_service.log_comment(comment)
            # Keep track of what's logged
            return comments_logged + comments_before
        return comments_logged

    def _log_remaining_comments(self, all_unique_comments: set, comments_logged: list) -> None:
        # Ensure last asset comments are filtered out
        all_comments_logged = set(comments_logged)
        still_comments_to_log = list(all_unique_comments - all_comments_logged)
        # Log remaining comments after last asset
        if still_comments_to_log:
            for comment in still_comments_to_log:
                if comment:
                    self._iteration_service.log_comment(comment)
        # Logging for comments captured
        all_comments_logged.discard(None)
        filtered_comments = list(filter(lambda comment: comment is not None, still_comments_to_log))
        if filtered_comments or all_comments_logged:
            _logger.info(f"Comments logged in iteration {self._iteration.name!r}.")

    def _log_modeva_asset(self, asset: TModel | TDataset | TVecticeObjects | TWidgetOutput):
        try:
            # if asset and (asset.get("models") or "dataframe" in asset):
            # self._iteration._log_fact_sheet(asset)  # type: ignore[reportPrivateUsage]
            if asset and "table" in asset:
                self._iteration._log_validation_result(asset)  # type: ignore[reportPrivateUsage]
        except Exception:
            pass

    def _log_asset_types(
        self,
        asset: TModel | TDataset | TVecticeObjects | TWidgetOutput,
        asset_type: VecticeType | GiskardType | ModevaType,
    ) -> None:
        if asset_type == VecticeType.VECTICE_OBJECT:
            self._log_vectice_asset(asset)  # type: ignore[reportArgumentType]
        elif asset_type == VecticeType.MODEL:
            self._log_model(asset)  # type: ignore[reportArgumentType]
        elif asset_type == VecticeType.DATASET:
            self._log_dataset(asset)  # type: ignore[reportArgumentType]
        elif asset_type == VecticeType.METRIC:
            self._log_metric_asset(asset)  # type: ignore[reportArgumentType]
        elif asset_type == GiskardType.RAG:
            self._iteration._log_rag_report(asset)  # type: ignore[reportPrivateUsage]
        elif asset_type == GiskardType.SCAN:
            self._iteration._log_scan_report(asset)  # type: ignore[reportPrivateUsage]
        elif asset_type == GiskardType.QATESTSET:
            self._iteration._log_qa_test(asset)  # type: ignore[reportPrivateUsage]
        elif asset_type == GiskardType.TEST_SUITE_RESULT:
            self._iteration._log_test_suite_result(asset)  # type: ignore[reportPrivateUsage]
        elif asset_type == ModevaType.VALIDATION_RESULT:
            self._log_modeva_asset(asset)

    def _log_assets(
        self,
        assets: list[TModel | TDataset | TVecticeObjects | TWidgetOutput],
    ):
        from vectice.services import iteration_service

        # all comments and variables
        all_comments_and_variables = self._get_all_cells_comments()
        # Unique comment set
        all_unique_comments = set([data["comment"] for data in all_comments_and_variables])
        # Keep track of what is logged
        comments_logged = []
        for asset in assets:
            asset_type = asset["asset_type"]
            try:
                comments_logged = self._log_comments_before_asset(all_comments_and_variables, asset, comments_logged)
                self._log_asset_types(asset, asset_type)
            except Exception as e:
                self._failed_assets.append({"reason": e, "asset": asset["variable"], "type": asset_type.value})
        # Log the remaining comments
        self._log_remaining_comments(all_unique_comments, comments_logged)
        # After logging, don't re-use code file
        iteration_service.lineage_file_id = None

    def _log_model(self, model: TModel):
        vectice_model = model["model"]
        self._iteration_service.log_model(vectice_model)
        _logger.info(f"Model {vectice_model.name!r} logged in iteration {self._iteration.name!r}.")
        model_summary = model.get("summary")

        if model_summary:
            # log statsmodel summary
            self._iteration_service.log_comment(model_summary)

    def _log_metric_asset(self, metric: TMetric):
        metric_tables = metric.get("tables")
        for table in metric_tables:
            if table is not None:
                self._iteration_service.log_table(table)
                _logger.info(f"Table {table.name!r} logged in iteration {self._iteration.name!r}.")

    def _log_vectice_asset(self, asset: TVecticeObjects) -> None:
        asset_to_log = asset["vectice_object"]
        if isinstance(asset_to_log, ValidationModel):
            object_name = asset["variable"]
        else:
            object_name = asset_to_log.name

        ensure_correct_project_id_from_representation_objs(self._iteration.project.id, asset_to_log)

        if isinstance(asset_to_log, Dataset):
            self._iteration_service.log_dataset(asset_to_log)
            _logger.info(f"Dataset {object_name!r} logged in iteration {self._iteration.name!r}.")
        elif isinstance(asset_to_log, Table):
            self._iteration_service.log_table(asset_to_log)
            _logger.info(f"Table {object_name!r} logged in iteration {self._iteration.index!r}.")
        elif isinstance(asset_to_log, Model):
            if asset_to_log.predictor:
                self._iteration._assign_model_predictor_metadata(asset_to_log)  # pyright: ignore[reportPrivateUsage]
            self._iteration_service.log_model(asset_to_log)
            _logger.info(f"Model {object_name!r} logged in iteration {self._iteration.index!r}.")
        elif isinstance(
            asset_to_log,
            (ModelRepresentation, ModelVersionRepresentation, DatasetRepresentation, DatasetVersionRepresentation),
        ):
            asset_type, already_assigned = self._iteration_service.assign_version_representation(
                asset_to_log
            )  # pyright: ignore[reportPrivateUsage]
            link = "already linked" if already_assigned else "linked"
            asset_name_variable = asset["variable"]
            is_parent = isinstance(asset_to_log, (ModelRepresentation, DatasetRepresentation))
            asset_version = (
                asset_to_log._last_version.name  # pyright: ignore [reportPrivateUsage]
                if (is_parent and asset_to_log._last_version is not None)  # pyright: ignore [reportPrivateUsage]
                else asset_to_log.name
            )
            asset_name = (
                asset_to_log.name
                if is_parent
                else _get_asset_parent_name(asset_to_log)  # pyright: ignore [reportPrivateUsage]
            )
            _logger.info(
                f"{asset_type} {asset_name!r} {asset_version} named {asset_name_variable!r} as a variable {link} to iteration {self._iteration.index!r}."
            )
        else:
            self._iteration._log_validation_model(asset_to_log)  # pyright: ignore[reportPrivateUsage]

    def _log_dataset(self, dataset: TDataset) -> None:
        vec_dataset = dataset["dataset"]
        try:
            self._iteration_service.log_dataset(vec_dataset)
        except FileNotFoundError:
            self._iteration_service.log_dataset(vec_dataset)
        _logger.info(f"Dataset {vec_dataset.name!r} logged in iteration {self._iteration.name!r}.")

    def _get_graphs(self, is_notebook: bool, capture_widget_graphs: bool = False) -> list[Any]:
        global GRAPHS
        if capture_widget_graphs and self._get_widget_graphs():
            # update graphs with widget graphs
            self.extract_and_save_widget_graphs()
        graphs = []
        if is_notebook:
            displayed_graphs = [displayed_graph for cell in GRAPHS.values() for displayed_graph in cell["displayed"]]
            saved_graphs = {graph for cell in GRAPHS.values() for graph in cell["saved"]}
        else:
            cell_id = _get_cell_id(self._ip)
            cells_graphs = GRAPHS[cell_id]
            displayed_graphs = cells_graphs["displayed"]
            saved_graphs = set(cells_graphs["saved"])
        graphs += list(saved_graphs)
        graphs += displayed_graphs
        return graphs

    def _get_widget_graphs(self) -> bool:
        """NB Only supports plotly currently."""
        from importlib.metadata import version

        from packaging.version import Version

        is_ipywidgets, is_kaleido = False, False
        if find_spec("ipywidgets") is not None:
            is_ipywidgets = True

        if is_ipywidgets and find_spec("kaleido") is not None:
            is_kaleido = True
            kaleido_version = version("kaleido")
            if Version(kaleido_version) >= Version("0.2.0"):
                _logger.warning(
                    f"Kaleido {kaleido_version!r} has known issues with plotly write_image. Install 0.1.0 or 0.1.0post1 for a stable experience."
                )

        if is_plotly:
            plotly_version = version("plotly")
            if Version(plotly_version) >= Version("6.0.0"):
                _logger.warning(
                    f"Plotly {plotly_version!r} and greater has known issues with ipywidgets. Install an older version of plotly for a stable experience."
                )

        if not (is_plotly and is_ipywidgets and is_kaleido):
            return False
        return True

    def extract_and_save_widget_graphs(self):
        """Extracts and saves figures from all detected `ipywidgets.Output` widgets or from the autolog.cell `ipywidgets.Output`.

        We have plotly write_img monkey patched so implicitly we are capturing what we write as an image below.
        """
        import plotly.graph_objects as go

        # Will be the cells widget or all widgets
        output_widgets = [
            asset["output"]  # pyright: ignore[reportGeneralTypeIssues]
            for asset in self._assets  # pyright: ignore[reportGeneralTypeIssues]
            if asset["asset_type"] == VecticeType.IPYWIDGETS_OUTPUT  # pyright: ignore[reportGeneralTypeIssues]
        ]
        for widget in output_widgets:
            for figure_count, out in enumerate(widget.outputs):
                is_plotly_graph = (
                    isinstance(out, dict) and "data" in out and "application/vnd.plotly.v1+json" in out["data"]
                )
                if is_plotly_graph:
                    fig_dict = out["data"]["application/vnd.plotly.v1+json"]
                    fig = go.Figure(fig_dict)  # pyright: ignore[reportCallIssue]
                    # Generate a meaningful filename based on the plot title
                    plot_title = (
                        fig.layout.title.text  # pyright: ignore[reportAttributeAccessIssue]
                        if fig.layout.title.text  # pyright: ignore[reportAttributeAccessIssue]
                        else f"figure_{figure_count}"  # pyright: ignore[reportAttributeAccessIssue]
                    )
                    invalid_chars_pattern = r'[ ,<>:"/\\|?*]'
                    filename = f"{re.sub(invalid_chars_pattern, '_', plot_title)}.png"
                    # Save the figure and trigger monkey patch
                    fig.write_image(filename)



from __future__ import annotations

import logging
from typing import BinaryIO

from vectice.api.http_error import HttpError
from vectice.api.rest_api import RestApi

_logger = logging.getLogger(__name__)


class ColumnsDescriptionApi(RestApi):
    def post_columns_description(self, id: str, file: tuple[str, tuple[str, BinaryIO]]):
        try:
            self._post_attachments(f"/api/columns-description/{id}", [file])
        except HttpError as e:
            self._httpErrorHandler.handle_post_http_error(e, "columns description", "update")



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Set

from vectice import Connection
from vectice.api.http_error_handlers import VecticeException
from vectice.connection import DEFAULT_HOST
from vectice.services.phase_service import PhaseService

_logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from vectice.api.client import Client
    from vectice.models.iteration import Iteration, IterationService
    from vectice.models.phase import Phase


class TraceClientService:
    """TraceClientService is a base class for services that interact with the Vectice Trace API.

    It provides functionality to initialize the service and manage the client connection.
    It also manages client configuration, including setting the API token, host, and phase.
    The intermediary logging calls for the Vectice Trace API are handled through this service aswell.
    """

    def __init__(
        self,
        phase: str | None = None,
        create_new_iteration: bool = False,
    ):
        """Initialize the ClientService with a Vectice client.

        :param client: An instance of the Vectice client.
        """
        api_token = self.get_env_api_token()
        host = self.get_env_host()

        self._connection = Connection(api_token=api_token, host=host) if api_token and host else None
        self._client = self._connection._client if self._connection else None  # pyright: ignore[reportPrivateUsage]
        self._phase = self._connection.phase(phase) if self._connection and phase else None

        self._iteration = self.get_active_iteration_or_create(create_new_iteration)
        self._iteration_service = (
            self._iteration._service if self._iteration else None  # pyright: ignore[reportPrivateUsage]
        )

    #### Properties ####

    @property
    def connection(self) -> Connection:
        """Get the Vectice connection."""
        if self._connection is None:
            raise VecticeException("Connection is not initialized. Please set the connection parameters first.")
        return self._connection

    @property
    def client(self) -> Client:
        """Get the Vectice client."""
        if self._client is None:
            raise VecticeException("Client is not initialized. Please set the connection parameters first.")
        return self._client

    @property
    def iteration(self) -> Iteration:
        """Get the current Vectice iteration."""
        if self._iteration is None:
            raise VecticeException("Iteration is not set. Please set the iteration first.")
        return self._iteration

    @property
    def iteration_service(self) -> IterationService:
        """Get the current Vectice iteration."""
        if self._iteration_service is None:
            raise VecticeException("Iteration is not set. Please set the iteration first.")
        return self._iteration_service

    @property
    def phase(self) -> Phase:
        """Get the current Vectice iteration."""
        if self._phase is None:
            raise VecticeException("Phase is not set. Please set the phase first.")
        return self._phase

    #### Utility Methods ####

    @staticmethod
    def get_env_api_token() -> str:
        api_token = os.environ.get("VECTICE_API_TOKEN")
        if not api_token:
            raise VecticeException(
                "`VECTICE_API_TOKEN` is not set. Please set the environmental variable for `VECTICE_API_TOKEN`"
            )
        return api_token

    @staticmethod
    def get_env_host() -> str:
        host = os.environ.get("VECTICE_HOST")
        if not host:
            _logger.warning("`VECTICE_HOST` is not set. Using default host: %s", DEFAULT_HOST)
            return DEFAULT_HOST
        return host

    def is_vectice_client_valid(self):
        """Check if the Vectice client configuration is valid."""
        if self._connection is None:
            raise VecticeException("Vectice connection is not initialized. Please set the connection parameters first.")
        if self._client is None:
            raise VecticeException("Vectice client is not initialized. Please set the connection parameters first.")
        if self._phase is None:
            raise VecticeException("Vectice phase is not set. Please set the phase first.")
        if self._iteration is None:
            raise VecticeException("Vectice iteration is not set. Please set the iteration first.")
        if self._iteration_service is None:
            raise VecticeException("Vectice iteration service is not initialized. Please set the iteration first.")
        return True

    #### Configuration ####

    def get_active_iteration_or_create(self, create_new_iteration: bool) -> Iteration | None:
        """Get the current iteration."""
        if self._phase is None:
            return None
        if create_new_iteration is True:
            iteration = self._phase.create_iteration()
        else:
            iteration = PhaseService(
                self._phase._client  # pyright: ignore[reportPrivateUsage]
            ).get_active_iteration_or_create(self._phase)
        return iteration

    def config(self, phase: str | None = None, create_new_iteration: bool | None = None) -> None:
        """Configure the Vectice trace service with API token, phase, host, and optional prefix."""
        if not self.connection:
            raise VecticeException(
                "Vectice connection is not initialized. Please set `VECTICE_API_TOKEN` and `VECTICE_HOST` environmental variables."
            )

        if phase:
            self._phase = self.connection.phase(phase)

        if create_new_iteration:
            self._iteration = self.get_active_iteration_or_create(create_new_iteration)
            self._iteration_service = self._iteration_service if self._iteration else None

    #### Logging Calls ####

    def save_autolog_assets(self, cells: Set[str], prefix: str | None = None, is_trace: bool = False) -> None:
        """Save autolog assets to the current iteration.

        :param cells: A list of dictionaries containing cell information.
        :param prefix: An optional prefix for the asset names.
        """
        if not cells:
            _logger.warning("No cells to save. Skipping trace assets saving.")
            return

        if prefix is None:
            prefix = self.phase.name
        formatted_cells = [{"cellId": str(cell_id), "content": cell} for cell_id, cell in enumerate(cells) if cell]
        self._iteration_service.save_autolog_assets(formatted_cells, prefix, is_trace)  # type: ignore



from __future__ import annotations

import logging
from io import BufferedReader, IOBase
from textwrap import dedent
from typing import TYPE_CHECKING

from PIL.Image import Image

from vectice.api.http_error_handlers import ArtifactVersionIdExistingError, VecticeException
from vectice.api.json import DatasetRegisterOutput, EntityFileOutput, ModelRegisterOutput
from vectice.api.json.iteration import (
    IterationStepArtifactEntityMetadataInput,
    IterationStepArtifactInput,
    IterationStepArtifactType,
    IterationUpdateInput,
)
from vectice.utils.common_utils import (
    check_string_sanity,
    get_image_or_file_variables,
    get_notebook_path,
    get_script_path,
    set_dataset_attachments,
    set_model_attachments,
)
from vectice.utils.dataframe_utils import transform_table_to_metadata_dict

if TYPE_CHECKING:
    from vectice.api.client import Client
    from vectice.api.json.code_version import CodeVersion
    from vectice.models.dataset import Dataset
    from vectice.models.iteration import Iteration
    from vectice.models.model import Model
    from vectice.models.representation.dataset_representation import DatasetRepresentation
    from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
    from vectice.models.representation.model_representation import ModelRepresentation
    from vectice.models.representation.model_version_representation import ModelVersionRepresentation
    from vectice.models.table import Table

    TAssetType = ModelRepresentation | ModelVersionRepresentation | DatasetRepresentation | DatasetVersionRepresentation

_logger = logging.getLogger(__name__)

lineage_file_id = None
code_source_file = get_notebook_path() or get_script_path()


class IterationService:
    def __init__(self, iteration: Iteration, client: Client):
        self._client = client
        self._iteration = iteration

    def log_image_or_file(self, asset: str | IOBase | Image, section: str | None = None):
        _, filename = self._create_image_or_file_artifact(asset, section)
        return filename

    def log_comment(self, asset: int | float | str, section: str | None = None):
        if isinstance(asset, str):
            check_string_sanity(asset)
        artifact_inp: list[IterationStepArtifactInput] = [IterationStepArtifactInput(type="Comment", text=str(asset))]
        self._client.add_iteration_artifacts(iteration_id=self._iteration.id, artifacts=artifact_inp, section=section)

    def log_table(self, asset: Table, section: str | None = None):
        data = transform_table_to_metadata_dict(asset)
        entity_metadata = IterationStepArtifactEntityMetadataInput(name=asset.name, content={"data": data})

        artifact_inp: list[IterationStepArtifactInput] = [
            IterationStepArtifactInput(type="EntityMetadata", entityMetadata=entity_metadata)
        ]
        self._client.add_iteration_artifacts(iteration_id=self._iteration.id, artifacts=artifact_inp, section=section)

    def log_model(self, asset: Model, section: str | None = None):
        from vectice import code_file_capture

        global code_source_file

        code_version = self._get_code_version()
        model_data = self._client.register_model(
            model=asset,
            phase=self._iteration.phase,
            iteration=self._iteration,
            code_version=code_version,
            section=section,
        )
        if code_file_capture:
            if code_source_file:
                self._attach_code_file_to_lineage(model_data, code_source_file)
        attachments_output, success_pickle = set_model_attachments(self._client, asset, model_data.model_version)
        return model_data, attachments_output, success_pickle

    def log_dataset(self, asset: Dataset, section: str | None = None):
        from vectice import code_file_capture

        global code_source_file

        code_version = self._get_code_version()
        dataset_output = self._client.register_dataset_from_source(
            dataset=asset,
            iteration_id=self._iteration.id,
            code_version=code_version,
            section=section,
        )
        if code_file_capture:
            if code_source_file:
                self._attach_code_file_to_lineage(dataset_output, code_source_file)
        attachments_output = set_dataset_attachments(self._client, asset, dataset_output.dataset_version)
        return dataset_output, attachments_output

    def assign_version_representation(
        self,
        asset: TAssetType,
        section: str | None = None,
    ) -> tuple[str, bool]:
        from vectice.models.representation.dataset_representation import DatasetRepresentation
        from vectice.models.representation.model_representation import ModelRepresentation

        asset_type = self._get_asset_type_from_asset_representation(asset)

        is_parent = isinstance(asset, (ModelRepresentation, DatasetRepresentation))
        asset_id = self._get_last_version_id(asset) if is_parent else asset.id
        artifact_inp = [
            IterationStepArtifactInput(
                type=asset_type,
                id=asset_id,
            )
        ]
        already_assigned: bool = False
        try:
            self._client.add_iteration_artifacts(
                iteration_id=self._iteration.id, artifacts=artifact_inp, section=section
            )
        except ArtifactVersionIdExistingError:
            already_assigned = True

        return asset_type, already_assigned

    def save_autolog_assets(self, cells: list[dict[str, str]], prefix: str | None = None, is_trace: bool = False):
        self._client.save_autolog_cells(self._iteration.id, cells, prefix, is_trace)

    def organize_with_ai(self):
        self._client.organize_with_ai(self._iteration.id)

    def organize_with_ai_and_report(self):
        report_data = self._client.organize_with_ai_and_report(self._iteration.id)
        hyper_link = f"{self._client.auth.api_base_url}/project/{self._iteration.phase.project.id}/reports"
        logging_output = dedent(
            f"""
                The report '{report_data.name}' is currently being generated and will be accessible via this link in a few minutes:
                {hyper_link}
            """
        ).lstrip()

        _logger.info(logging_output)

    def _get_asset_type_from_asset_representation(
        self,
        asset: TAssetType,
    ):
        from vectice.models.representation.dataset_representation import DatasetRepresentation
        from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
        from vectice.models.representation.model_representation import ModelRepresentation
        from vectice.models.representation.model_version_representation import ModelVersionRepresentation

        if isinstance(asset, (ModelRepresentation, ModelVersionRepresentation)):
            return "ModelVersion"
        if isinstance(
            asset, (DatasetRepresentation, DatasetVersionRepresentation)
        ):  # pyright: ignore[reportUnnecessaryIsInstance]
            return "DataSetVersion"

        raise ValueError(f"Asset type '{type(asset)!s}' is unsupported")

    def _get_last_version_id(self, value: ModelRepresentation | DatasetRepresentation):
        last_version = value._last_version  # pyright: ignore [reportPrivateUsage]
        if last_version is None:
            raise VecticeException("Unable to log asset, no versions exist.")
        return last_version.id

    def update(self, name: str | None = None, description: str | None = None):
        if name is None and description is None:
            return

        self._client.update_iteration(self._iteration.id, IterationUpdateInput(name=name, description=description))
        if name is not None:
            self._iteration.name = name

        self._iteration.description = description
        _logger.info(f"Iteration {self._iteration.id!r} successfully updated.")

    def _create_code_version_file_artifact(
        self, value: str | IOBase | Image, lineage_id: int
    ) -> EntityFileOutput | None:
        is_image_or_file, filename = get_image_or_file_variables(value)
        try:
            artifact, *_ = self._client.upsert_lineage_attachments(
                files=[("file", (filename, is_image_or_file))], lineage_id=lineage_id
            )
            if isinstance(is_image_or_file, BufferedReader):
                is_image_or_file.close()
            return EntityFileOutput(
                fileId=artifact.fileId,
                fileName=artifact.fileName,
                contentType=artifact.contentType,
                entityId=artifact.entityId,
                entityType=artifact.entityType,
            )
        except VecticeException as error:
            raise error
        except Exception:
            pass

    def _create_image_or_file_artifact(
        self, value: str | IOBase | Image, section: str | None = None
    ) -> tuple[IterationStepArtifactInput, str]:
        is_image_or_file, filename = get_image_or_file_variables(value)
        try:
            artifact, *_ = self._client.upsert_iteration_attachments(
                files=[("file", (filename, is_image_or_file))], iteration_id=self._iteration.id, step_name=section
            )
            if isinstance(is_image_or_file, BufferedReader):
                is_image_or_file.close()
            return (
                IterationStepArtifactInput(id=artifact.fileId, type=IterationStepArtifactType.EntityFile.name),
                filename,
            )
        except VecticeException as error:
            raise error
        except Exception as error:
            raise ValueError("Check the provided image.") from error

    def _get_code_version(self) -> CodeVersion | None:
        # TODO: cyclic imports
        from vectice import code_capture
        from vectice.models.code_version import capture_code_version, inform_if_git_repo

        if code_capture:
            return capture_code_version()
        else:
            inform_if_git_repo()
            return None

    def _attach_code_file_to_lineage(
        self, asset_output: DatasetRegisterOutput | ModelRegisterOutput, code_source_file: str
    ) -> None:
        global lineage_file_id

        try:
            lineage_id = None
            if hasattr(asset_output, "dataset_version"):
                lineage_id = asset_output.dataset_version.origin_id  # pyright: ignore[reportAttributeAccessIssue]
            if hasattr(asset_output, "model_version"):
                lineage_id = asset_output.model_version.origin_id  # pyright: ignore[reportAttributeAccessIssue]

            if lineage_id and lineage_file_id:
                self._client.attach_file_to_lineage(lineage_id, lineage_file_id)
            if lineage_id and lineage_file_id is None:
                attachment_output = self._create_code_version_file_artifact(code_source_file, lineage_id)
                lineage_file_id = attachment_output.fileId if attachment_output else None
        except Exception:
            pass



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pyspark.sql import SparkSession
    from pyspark.sql.dataframe import DataFrame


class TableSparkSession(metaclass=ABCMeta):
    """Base class for TableSparkSession."""

    @abstractmethod
    def __init__(
        self,
        spark_client: SparkSession,
    ):
        """Initialize a resource."""
        self._spark_client = spark_client
        self._table_path: str | None = None

    @property
    def table_path(self) -> str | None:
        """The session's table path.

        Returns:
            The session's table path.
        """
        return self._table_path

    @table_path.setter
    def table_path(self, value: str):
        """Set session's table path.

        Parameters:
            value: The session's table path to set.
        """
        self._table_path = value

    @abstractmethod
    def set_table_key(self, table_name: str) -> None:
        pass

    @abstractmethod
    def set_rows_table_name(self, key_rows_table_name: str) -> None:
        pass

    @abstractmethod
    def get_detail(self) -> list:
        pass

    @abstractmethod
    def get_detail_as_dict(self) -> dict[str, Any]:
        pass

    @abstractmethod
    def get_rows(self) -> list:
        pass

    @abstractmethod
    def get_number_of_rows(self) -> int:
        pass

    @abstractmethod
    def get_schema(self) -> list:
        pass

    @abstractmethod
    def read_table_path(self, table_name: str) -> DataFrame:
        pass

    @abstractmethod
    def get_table_extended(self) -> DataFrame:
        pass

    @abstractmethod
    def get_table_history(self) -> list:
        pass



from __future__ import annotations

from vectice.models.additional_info import AdditionalInfo, ExtraInfo, Framework
from vectice.models.attachment_container import AttachmentContainer
from vectice.models.errors import VecticeError
from vectice.models.iteration import Iteration
from vectice.models.metric import Metric
from vectice.models.phase import Phase
from vectice.models.project import Project
from vectice.models.property import Property
from vectice.models.workspace import Workspace

__all__ = [
    "AttachmentContainer",
    "Metric",
    "Property",
    "Project",
    "VecticeError",
    "Workspace",
    "Phase",
    "Iteration",
    "AdditionalInfo",
    "ExtraInfo",
    "Framework",
]



from __future__ import annotations

from typing import Any, Dict

from vectice.api.json.review import ReviewOutput
from vectice.utils.common_utils import (
    repr_class,
)


class ReviewRepresentation:
    """Represents the metadata of a Vectice review.

    A Review Representation shows information about a specific review from the Vectice app.
    It makes it easier to get and read this information through the API.

    Attributes:
        id (str): The unique identifier of the review.
        message (str | None): The message of the review.
        feedback (str | None): The feedback of the review.
        status (str): The status of the review (In Review, Approved, Rejected, or Cancelled).
        creator (Dict[str, str]): Creator of the review.
        assignee (Dict[str, str]): Assignee of the review.
    """

    def __init__(self, output: ReviewOutput):
        self.id = output.id
        self.message = output.message
        self.status = output.status
        self.feedback = output.feedback
        self.created_date = output.created_date
        self.creator = output.creator
        self.assignee = output.assignee

    def __repr__(self):
        return repr_class(self)

    def asdict(self) -> Dict[str, Any]:
        """Transform the ReviewRepresentation into a organised dictionary.

        Returns:
            The object represented as a dictionary
        """
        return {
            "id": self.id,
            "message": self.message,
            "feedback": self.feedback,
            "status": self.status,
            "created_date": self.created_date,
            "creator": self.creator,
            "assignee": self.assignee,
        }



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Hashable

if TYPE_CHECKING:
    from pandas import DataFrame

    from vectice.models.table import Table

_logger = logging.getLogger(__name__)


def repr_list_as_pd_dataframe(list_arg: list | dict) -> DataFrame:  # type: ignore
    try:
        import pandas as pd
    except Exception as err:
        raise ModuleNotFoundError("To use this method, please install pandas first") from err
    if isinstance(list_arg, dict) and list_arg.get("data"):
        return pd.DataFrame(list_arg["data"])
    return pd.DataFrame(list_arg)


def _convert_datetime_columns_to_str(dataframe: DataFrame) -> DataFrame:
    datetime_columns = dataframe.select_dtypes(include=["datetime"]).columns
    if len(datetime_columns) >= 1:
        dataframe[datetime_columns] = dataframe[datetime_columns].astype(str)
    return dataframe


def transform_table_to_metadata_dict(asset: Table) -> list[dict[Hashable, Any]]:
    dataframe_reset_index = _reset_index(asset.dataframe).fillna("NaN")
    dataframe_cleaned = _convert_datetime_columns_to_str(dataframe_reset_index)
    try:
        import pandas as pd

        # convert objects to strings so the dataframe is json serializable
        dataframe_cleaned[dataframe_cleaned.select_dtypes(["object"]).columns] = dataframe_cleaned.select_dtypes(
            ["object"]
        ).astype(pd.StringDtype())
    except Exception:
        pass
    row_count, col_count = dataframe_cleaned.shape
    max_row_count = 100
    if row_count > max_row_count:
        dataframe_cleaned = dataframe_cleaned.iloc[0:max_row_count, :]
        _logger.warning(
            f"Only first {max_row_count} rows were captured. For additional capacity, please contact your sales representative or email support@vectice.com"
        )
    max_col_count = 20
    if col_count > max_col_count:
        dataframe_cleaned = dataframe_cleaned.iloc[:, 0:max_col_count]
        _logger.warning(
            f"Only first {max_col_count} columns were captured. For additional capacity, please contact your sales representative or email support@vectice.com"
        )
    columns_content = dataframe_cleaned.to_dict("records")
    return columns_content


def _reset_index(df: DataFrame) -> DataFrame:
    if not df.index.equals(df.index.to_frame().reset_index(drop=True).index):
        return df.reset_index()
    else:
        return df



from __future__ import annotations

import logging
from typing import Any

from vectice.autolog.asset_services.service_types.vectice_types import VecticeType
from vectice.models.table import Table
from vectice.utils.common_utils import get_asset_name

_logger = logging.getLogger(__name__)


class AutologH2oDisplayService:
    def __init__(
        self,
        key: str,
        asset: Any,
        phase_name: str,
        prefix: str | None = None,
    ):
        self._asset = asset
        self._key = key
        self._display_name = get_asset_name(self._key, phase_name, prefix)

    def get_asset(self):
        try:
            dataframe = self._asset.as_data_frame()
            table = Table(dataframe=dataframe, name=self._display_name)

            return {"variable": self._key, "vectice_object": table, "asset_type": VecticeType.VECTICE_OBJECT}
        except Exception as e:
            _logger.debug(f"Failed to get asset from {self._asset.__class__.__name__}: {e!s}.")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List

from vectice.api._utils import read_nodejs_date
from vectice.api.json.json_type import TJSON

if TYPE_CHECKING:
    from datetime import datetime

    from vectice.api.json.dataset_representation import DatasetRepresentationOutput


class DatasetVersionRepresentationOutput(TJSON):
    @property
    def created_date(self) -> datetime | None:
        return read_nodejs_date(str(self["createdDate"]))

    @property
    def updated_date(self) -> datetime | None:
        return read_nodejs_date(str(self["updatedDate"]))

    @property
    def id(self) -> str:
        return str(self["vecticeId"])

    @property
    def dataset_id(self) -> str | None:
        if "dataSet" in self:
            return str(self["dataset"]["vecticeId"])
        return None

    @property
    def phase_origin(self) -> str | None:
        origins = self.get("origins", {})
        phase = origins.get("phase")
        if phase and "vecticeId" in phase:
            return str(phase["vecticeId"])
        return None

    @property
    def iteration_origin(self) -> str | None:
        origins = self.get("origins", {})
        iteration = origins.get("iteration")
        if iteration and "vecticeId" in iteration:
            return str(iteration["vecticeId"])
        return None

    @property
    def creator(self) -> dict[str, str]:
        return {"name": self["createdBy"]["name"], "email": self["createdBy"]["email"]}

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def origin(self) -> str:
        return str(self["sourceOrigin"])

    @property
    def dataset_name(self) -> str | None:
        if "dataSet" in self:
            return str(self["dataSet"]["name"])
        return None

    @property
    def dataset(self) -> DatasetRepresentationOutput:
        from vectice.api.json.dataset_representation import DatasetRepresentationOutput

        return DatasetRepresentationOutput(**self["dataSet"])

    @property
    def description(self) -> str | None:
        return str(self["description"]) if self["description"] else None

    @property
    def project_id(self) -> str | None:
        if "dataSet" in self:
            return str(self["dataSet"]["project"]["vecticeId"])
        return None

    @property
    def properties(self) -> List[Dict[str, Any]]:
        return self["properties"]

    @property
    def resources(self) -> List[Dict[str, Any]]:
        for source in self["datasetSources"]:
            source["usage"] = "GENERIC" if source["usage"] is None else source["usage"]
            source["resource_type"] = source.pop("usage")
            source["number_of_items"] = source.pop("itemsCount")
            source["total_number_of_columns"] = source.pop("columnsCount")

        return self["datasetSources"]

    @property
    def is_starred(self) -> bool:
        return bool(self["isStarred"])



from __future__ import annotations

from dataclasses import InitVar, dataclass, field
from functools import reduce
from typing import Generic, Type, TypeVar

from vectice.api.json.json_type import TJSON
from vectice.api.json.page import Page

ItemType = TypeVar("ItemType")


@dataclass
class PagedResponse(Generic[ItemType]):
    """Generic structure describing a result of a paged request.

    The structure contains page information and the list of items for this page.
    """

    total: int
    """Total number of available pages."""
    list: list[ItemType] = field(init=False)
    """Current list of elements for this page."""
    current_page: Page = field(init=False)
    """Information on the current page."""
    page: InitVar[TJSON]
    item_cls: InitVar[Type[ItemType]]
    items: InitVar[list[TJSON]]

    def __post_init__(self, page: TJSON, cls: Type[ItemType], items: list[TJSON]):
        self.current_page = Page(**page)
        self.list = reduce(lambda acc, curr: [*acc, cls(**curr)], items, [])



from __future__ import annotations

from typing import Literal

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser
from vectice.api.gql_model import RETURNS_MODEL_VERSION
from vectice.api.json.paged_response import PagedResponse
from vectice.api.json.report import ReportOutput
from vectice.utils.api_utils import DEFAULT_NUMBER_OF_ITEMS, PAGINATE_OUTPUT, get_page_input

_RETURNS = f"""
    id
    name
    createdDate
    updatedDate
    createdBy {{
        name
        email
    }}
    modelVersion {{
        {RETURNS_MODEL_VERSION}
        __typename
    }}
    __typename
"""


class GqlReportApi(GqlApi):
    def get_reports(
        self, id: str, type: Literal["prj"] | Literal["mdv"], size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[ReportOutput]:
        gql_query = "CDTReports"
        variable_types = "$type:CDTReportQueryType,$filters:CDTReportFiltersInput!,$page:PageInput"
        kw = "type:$type,filters:$filters,page:$page"
        kw_name = "projectId" if type == "prj" else "modelVersionId"
        filters = {kw_name: id}
        variables = {
            "type": "PROJECT" if type == "prj" else "MODEL_VERSION",
            "filters": filters,
            "page": get_page_input(size=size),
        }
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=PAGINATE_OUTPUT.format(_RETURNS),
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            return Parser().parse_paged_response(response[gql_query])
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "model_version" if type == "mdv" else "project", id)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from vectice.autolog.asset_services.metric_service import MetricService
from vectice.autolog.asset_services.service_types.vectice_types import VecticeType
from vectice.autolog.asset_services.technique_service import TechniqueService
from vectice.autolog.model_library import ModelLibrary
from vectice.models.model import Model
from vectice.utils.common_utils import get_asset_name, temp_directory

if TYPE_CHECKING:
    from keras import Model as KerasModel  # type: ignore[reportMissingImports]  # type: ignore[reportMissingImports]
    from keras.layers import InputLayer  # type: ignore[reportMissingImports]
    from keras.models import Model as KerasModel  # type: ignore[reportMissingImports]


KERAS_TEMP_DIR = "keras"

_logger = logging.getLogger(__name__)


class AutologKerasService(MetricService, TechniqueService):
    def __init__(
        self,
        key: str,
        asset: KerasModel,
        data: dict,
        custom_metrics_data: set[str | None],
        phase_name: str,
        prefix: str | None = None,
    ):
        self._asset = asset
        self._key = key
        self._temp_dir = temp_directory(KERAS_TEMP_DIR)
        self._model_name = get_asset_name(self._key, phase_name, prefix)

        super().__init__(cell_data=data, custom_metrics=custom_metrics_data)

    def get_asset(self):
        model_metrics = self._get_model_metrics(self._cell_data)
        training_metrics = self._get_keras_training_metrics(self._asset)
        temp_file_path = self._get_keras_graph()
        _, params = self._get_keras_info(self._asset)

        model = Model(
            library=ModelLibrary.KERAS.value,
            technique=self._get_model_technique(self._asset, ModelLibrary.KERAS),
            metrics=model_metrics.update(training_metrics),
            properties=params,
            name=self._model_name,
            predictor=self._asset,
            attachments=temp_file_path,
        )
        return {
            "variable": self._key,
            "model": model,
            "asset_type": VecticeType.MODEL,
        }

    def _format_keras_params(self, model: KerasModel) -> dict[str, Any]:
        params: dict[str, Any] = {}

        def _get_output_shape(layer: InputLayer) -> tuple:
            try:
                # Keras 2.15.1 & below
                return layer.output_shape
            except AttributeError:
                return layer.output.shape  # pyright: ignore[reportAttributeAccessIssue]

        for i, layer in enumerate(model.layers):
            output_shape = _get_output_shape(layer)
            params[f"Layer-{i}"] = {
                "name": layer.name,
                "param": layer.count_params(),
                "output shape": output_shape,
            }
        params["Total # of weights"] = model.count_params()

        return params

    def _get_keras_info(self, model: KerasModel) -> tuple[str, dict[str, Any]] | tuple[None, None]:
        try:
            return "keras", self._format_keras_params(model)
        except Exception:
            return None, None

    def _get_keras_graph(self) -> str | None:
        graph = None
        try:
            from keras.utils import plot_model  # type: ignore[reportMissingImports]

            json_file_name = self._temp_dir / f"{self._key}_plot.png"
            temp_file_path = json_file_name.as_posix()
            graph = plot_model(self._asset, to_file=temp_file_path, show_shapes=True, show_layer_names=False)

            if graph is None:
                temp_file_path = None
                _logger.info(
                    "Unable to generate the model plot. Please check the 'plot_model' function from the Keras library is working correctly. Make sure that the graphviz and pydot packages are installed and configured properly."
                )
            return temp_file_path

        except Exception as e:
            _logger.info(
                f"Unable to generate the model plot. Ensure that the Keras library is correctly installed and up-to-date. Error details: {e}"
            )
        return None



from __future__ import annotations

from typing import Any, Dict

# TODO - This type can be improved, see : https://stackoverflow.com/questions/76345697/type-hinting-json-object-rigorously
TJSON = Dict[str, Any]



from __future__ import annotations

import logging
from functools import reduce
from typing import TYPE_CHECKING, Any, Dict

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser
from vectice.api.json.dataset_representation import DatasetRepresentationOutput
from vectice.api.json.dataset_resource_representation import DatasetResourceRepresentationOutput
from vectice.api.json.dataset_version_representation import DatasetVersionRepresentationOutput
from vectice.api.json.paged_response import PagedResponse
from vectice.models.resource.metadata.base import DatasetSourceType, DatasetSourceUsage
from vectice.utils.api_utils import DEFAULT_NUMBER_OF_ITEMS, PAGINATE_OUTPUT, get_page_input

if TYPE_CHECKING:
    from vectice.api.json.dataset_register import DatasetRegisterInput, DatasetRegisterOutput
    from vectice.api.json.iteration import IterationContextInput


_logger = logging.getLogger(__name__)

# TODO JobRun for lineages
_RETURNS = """
            datasetVersion {
                          vecticeId
                          name
                          origin {
                            id
                          }
                          dataSet {
                            name
                            projectId
                          }


            }
            useExistingDataset
            useExistingVersion
            __typename
            """

_RETURNS_DATASET = """
            vecticeId
            createdDate
            updatedDate
            name
            description
            type
            sourceOrigin
            versionCount
            lastVersion {
                vecticeId
                createdDate
                updatedDate
                name
                description
                __typename
            }
            project {
                vecticeId
            }
            __typename
"""

_BASE_DATASET_VERSION = """
            vecticeId
            createdDate
            updatedDate
            isStarred
            name
            description
            sourceOrigin
            createdBy {
                name
                email
            }
            properties {
                key
                value
            }
            __typename
"""

_LIST_VERSION = f"""
            {_BASE_DATASET_VERSION}
            origins {{
                phase {{
                    vecticeId
                    __typename
                }}
                iteration {{
                    vecticeId
                    __typename
                }}
                __typename
            }}
            datasetSources {{
                size
                usage
                itemsCount
                columnsCount
            }}
"""

RETURNS_DATASET_VERSION = f"""
            {_BASE_DATASET_VERSION}
            origins {{
                phase {{
                    vecticeId
                    __typename
                }}
                iteration {{
                    vecticeId
                    __typename
                }}
                __typename
            }}
            dataSet {{
                vecticeId
                name
                project {{
                    vecticeId
                }}
                sourceOrigin
                type
                description
                versionCount
                lastVersion {{
                    vecticeId
                    __typename
                }}
            }}
            datasetSources {{
                usage
                itemsCount
                size
                columnsCount
            }}
"""


_RETURNS_EMPTY_DATASET_VERSION = """
            vecticeId
            __typename
"""


_RETURNS_DATASET_SOURCES = """
        id
        type
        usage
        files {
            items {
                id
                name
                filename
                size
                rowsNumber
                columnsNumber
                uri
                extraMetadata {
                    key
                    value
                    displayName
                }
                __typename
            }
        }
        tables {
            items {
                id
                name
                tablename
                size
                rowsNumber
                columnsNumber
                uri
                extraMetadata {
                    key
                    value
                    displayName
                }
                __typename
            }
        }
        __typename
"""

_RETURNS_COLUMNS = PAGINATE_OUTPUT.format(
    """
    name
    categoryType
    description
    statistics {
        numerical {
            mean
            stdDeviation
            quantiles {
                qMin
                q25
                q50
                q75
                qMax
            }
            missing
        }
        text {
            unique
            missing
            mostCommons {
                value
                frequency
            }
        }
        date {
            minimum
            maximum
            median
            mean
            missing
        }
        boolean {
            true
            false
            missing
        }
    }
"""
)

_RETURNS_PAGINATED_DATASET_LIST = PAGINATE_OUTPUT.format(_RETURNS_DATASET)
_RETURNS_PAGINATED_VERSION_LIST = PAGINATE_OUTPUT.format(_LIST_VERSION)


class GqlDatasetApi(GqlApi):
    def get_dataset(self, id: str) -> DatasetRepresentationOutput:
        variable_types = "$datasetId:VecticeId!"
        kw = "datasetId:$datasetId"
        variables = {"datasetId": id}
        query = GqlApi.build_query(
            gql_query="getDataset",
            variable_types=variable_types,
            returns=_RETURNS_DATASET,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            dataset_output: DatasetRepresentationOutput = Parser().parse_item(response["getDataset"])
            return dataset_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset", id)

    def get_dataset_list(
        self, project_id: str, size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[DatasetRepresentationOutput]:
        variable_types = "$projectId:VecticeId!,$page:PageInput"
        kw = "projectId:$projectId,page:$page"
        variables = {
            "projectId": project_id,
            "page": get_page_input(size=size),
        }
        query = GqlApi.build_query(
            gql_query="getDatasetList",
            variable_types=variable_types,
            returns=_RETURNS_PAGINATED_DATASET_LIST,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            dataset_output: PagedResponse[DatasetRepresentationOutput] = Parser().parse_paged_response(
                response["getDatasetList"]
            )
            return dataset_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "project", project_id)

    def get_dataset_version_list(
        self, id: str, size: int = DEFAULT_NUMBER_OF_ITEMS
    ) -> PagedResponse[DatasetVersionRepresentationOutput]:
        gql_query = "getDatasetVersionsList"
        variable_types = "$datasetId:VecticeId!,$page:PageInput"
        kw = "datasetId:$datasetId,page:$page"
        variables = {
            "datasetId": id,
            "page": get_page_input(size=size),
        }
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS_PAGINATED_VERSION_LIST,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            dataset_output: PagedResponse[DatasetVersionRepresentationOutput] = Parser().parse_paged_response(
                response[gql_query]
            )
            return dataset_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset_version", id)

    def get_dataset_version(self, id: str) -> DatasetVersionRepresentationOutput:
        variable_types = "$datasetVersionId:VecticeId!"
        kw = "datasetVersionId:$datasetVersionId"
        variables = {"datasetVersionId": id}
        query = GqlApi.build_query(
            gql_query="getDatasetVersion",
            variable_types=variable_types,
            returns=RETURNS_DATASET_VERSION,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            dataset_output: DatasetVersionRepresentationOutput = Parser().parse_item(response["getDatasetVersion"])
            return dataset_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset_version", id)

    def get_dataset_version_has_columns_without_description(self, id: str) -> bool:
        variable_types = "$datasetVersionId:VecticeId!"
        kw = "datasetVersionId:$datasetVersionId"
        variables = {"datasetVersionId": id}
        query = GqlApi.build_query(
            gql_query="getDatasetVersion",
            variable_types=variable_types,
            returns="hasColumnWithoutDescription",
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            return self.execute(query_built, variables)["getDatasetVersion"]["hasColumnWithoutDescription"]
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset_version", id)

    def register_dataset(
        self,
        data: DatasetRegisterInput,
        iteration_context: IterationContextInput,
    ) -> DatasetRegisterOutput:
        variables: dict[str, Any] = {"iterationContext": iteration_context, "data": data}
        kw = "iterationContext:$iterationContext,data:$data"
        variable_types = "$iterationContext:IterationContextInput!,$data:DatasetRegisterInput!"

        query = GqlApi.build_query(
            gql_query="registerDataset",
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            dataset_output: DatasetRegisterOutput = Parser().parse_item(response["registerDataset"])
            return dataset_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset", "register dataset")

    def update_dataset_version(self, id: str, columns_description: list[dict[str, str]]):
        variable_types = "$datasetVersionId:VecticeId!,$data:DatasetVersionUpdateInput!"
        kw = "datasetVersionId:$datasetVersionId,data:$data"
        variables = {"datasetVersionId": id, "data": {"columnsDescription": columns_description}}
        query = GqlApi.build_query(
            gql_query="updateDatasetVersion",
            variable_types=variable_types,
            returns=_RETURNS_EMPTY_DATASET_VERSION,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            self.execute(query_built, variables)
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset_version", id)

    def has_column(self, id: str, column: str) -> bool:
        variable_types = "$datasetVersionId:VecticeId!,$column:String!"
        kw = "datasetVersionId:$datasetVersionId,column:$column"
        variables = {"datasetVersionId": id, "column": column}
        query = GqlApi.build_query(
            gql_query="hasColumn",
            variable_types=variable_types,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            return self.execute(query_built, variables)["hasColumn"]
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset_version", id)

    def get_resources_items(
        self, id: str, type: DatasetSourceUsage | None
    ) -> list[DatasetResourceRepresentationOutput]:
        gql_query = "resourcesItems"
        variable_types = "$datasetVersionId:VecticeId!,$resourceType:DatasetSourceUsage"
        kw = "datasetVersionId:$datasetVersionId,resourceType:$resourceType"
        variables = {
            "datasetVersionId": id,
            "resourceType": type,
        }
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS_DATASET_SOURCES,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response: Dict[str, Any] = self.execute(query_built, variables)
            dataset_output: list[DatasetResourceRepresentationOutput] = Parser().parse_list(
                reduce(
                    lambda acc, curr: acc
                    + [
                        {
                            **item,
                            "type": curr["type"],
                            "usage": curr.get("usage", None),
                        }
                        for item in (curr["files"]["items"] if curr["type"] == "FILES" else curr["tables"]["items"])
                    ],
                    response[gql_query],
                    [],
                )
            )
            return dataset_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset_sources", id)

    def get_columns(self, id: int, type: DatasetSourceType) -> list[Dict[str, Any]]:
        gql_query = "sourceColumns"
        variable_types = "$id:Float!,$columnSourceType:ColumnSourceType!"
        kw = "id:$id,columnSourceType:$columnSourceType"
        variables = {
            "id": id,
            "columnSourceType": "DBColumnSource" if type == DatasetSourceType.DB else "FileSourceColumn",
        }
        query = GqlApi.build_query(
            gql_query=gql_query,
            variable_types=variable_types,
            returns=_RETURNS_COLUMNS,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            return response[gql_query]["items"]
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "dataset_sources", id)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from vectice.models.resource.databricks.base import TableSparkSession

if TYPE_CHECKING:
    from pyspark.sql import SparkSession
    from pyspark.sql.dataframe import DataFrame


class DatabricksTableSparkSessionSQL(TableSparkSession):
    def __init__(self, spark_client: SparkSession):
        super().__init__(spark_client=spark_client)

    def set_table_key(self, table_name: str) -> None:
        self._spark_client.sql(f"SET key_table_name={table_name}")

    def set_rows_table_name(self, key_rows_table_name: str) -> None:
        self._spark_client.sql(f"SET key_rows_table_name={key_rows_table_name}")

    def get_detail(self) -> list:
        return self._spark_client.sql("DESCRIBE `${key_table_name}`").collect()

    def get_detail_as_dict(self) -> dict[str, Any]:
        return self._spark_client.sql("DESCRIBE DETAIL `${key_table_name}`").collect()[0].asDict()

    def get_rows(self) -> list:
        return self._spark_client.sql("DESCRIBE EXTENDED `${key_table_name}`").collect()

    def get_number_of_rows(self) -> int:
        return self._spark_client.sql("SELECT COUNT(*) FROM ${key_rows_table_name}").collect()[0][0]

    def get_schema(self) -> list:
        return self._spark_client.sql("DESCRIBE `${key_table_name}`").collect()

    def read_table_path(self, table_name: str) -> DataFrame:
        return self._spark_client.read.load(table_name)

    def get_table_history(self) -> list:
        return self._spark_client.sql("DESCRIBE HISTORY `${key_table_name}`").collect()

    def get_table_extended(self) -> DataFrame:
        return self._spark_client.sql("SHOW TABLE EXTENDED LIKE '${key_table_name}'")



from __future__ import annotations

import logging
from textwrap import dedent
from typing import TYPE_CHECKING

from rich.table import Table

from vectice.models.project import Project
from vectice.models.representation.project_template import ProjectTemplate
from vectice.utils.api_utils import DEFAULT_MAX_SIZE, DEFAULT_NUMBER_OF_ITEMS, DEFAULT_PRINT_SIZE
from vectice.utils.common_utils import temp_print
from vectice.utils.logging_utils import format_description

if TYPE_CHECKING:
    from vectice import Connection
    from vectice.api import Client


_logger = logging.getLogger(__name__)


class Workspace:
    """Represent a Vectice Workspace.

    Workspaces are containers used to organize projects, assets, and
    users.

    Vectice users have access to a personal workspace and other
    workspaces so they can learn and collaborate with other users. An
    organization will have many workspaces, each with an Admin and
    Members with different privileges.

    Note that only an Org Admin can create a new workspace in the
    organization.

    You can get a workspace from your [`Connection`][vectice.Connection]
    object by calling `workspace()`:

    ```python
    import vectice

    connect = vectice.connect(...)
    workspace = connect.workspace("Iris workspace")
    ```

    Or you can get it directly when connecting to Vectice:

    ```python
    import vectice

    workspace = vectice.connect(..., workspace="Iris workspace")
    ```

    See [`Connection.connect`][vectice.Connection.connect] to learn
    how to connect to Vectice.
    """

    def __init__(self, id: str, name: str, description: str | None = None):
        self._id = id
        self._name = name
        self._description = description
        self._client: Client
        self._connection: Connection

    def __post_init__(self, client: Client, connection: Connection):
        self._client = client
        self._connection = connection

    def __eq__(self, other: object):
        if not isinstance(other, Workspace):
            return NotImplemented
        return self.id == other.id

    def __repr__(self):
        description = self._description if self._description else "None"
        return f"Workspace(name={self.name!r}, id={self._id}, description={description!r})"

    @property
    def id(self) -> str:
        """The workspace's id.

        Returns:
            The workspace's id.
        """
        return self._id

    @property
    def name(self) -> str:
        """The workspace's name.

        Returns:
            The workspace's name.
        """
        return self._name

    @property
    def description(self) -> str | None:
        """The workspace's description.

        Returns:
            The workspace's description.
        """
        return self._description

    @property
    def properties(self) -> dict:
        """The workspace's name and id.

        Returns:
            A dictionary containing the `name` and `id` items.
        """
        return {"name": self.name, "id": self.id}

    def project(self, project: str) -> Project:
        """Get a project.

        Parameters:
            project: The project name or id.

        Returns:
            The project.
        """
        item = self._client.get_project(project, self.id)
        logging_output = dedent(
            f"""
                Project {item.name!r} successfully retrieved.

                For quick access to the Project in the Vectice web app, visit:
                {self._client.auth.api_base_url}/browse/project/{item.id}"""
        ).lstrip()
        _logger.info(logging_output)
        project_object = Project(item.id, self, item.name, item.description)
        return project_object

    def list_projects(
        self, number_of_items: int = DEFAULT_NUMBER_OF_ITEMS, display_print: bool = True
    ) -> list[Project]:
        """Retrieves a list of projects belonging to the workspace. It will also print the first 10 projects as a tabular form.

        Parameters:
            number_of_items: The number of datasets to retrieve. Defaults to 30.
            display_print: If set to True, it will print the first 10 projects in a tabular form.

        Returns:
            A list of `Project` instances.
        """
        if number_of_items > DEFAULT_MAX_SIZE:
            _logger.warning(
                "Only the first 100 projects will be retrieved. For additional projects, please contact your sales representative or email support@vectice.com"
            )
            number_of_items = DEFAULT_MAX_SIZE
        project_outputs = self._client.list_projects(self.id, size=number_of_items)
        project_list = [
            Project(project.id, self, project.name, project.description) for project in project_outputs.list
        ]

        if display_print:
            rich_table = Table(expand=True, show_edge=False)

            rich_table.add_column("Name", justify="left", no_wrap=True, max_width=15)
            rich_table.add_column("Description", justify="left", no_wrap=True, max_width=50)

            for project in project_outputs.list[:DEFAULT_PRINT_SIZE]:
                rich_table.add_row(project.id, project.name, format_description(project.description))

            description = dedent(
                f"""
            There are {project_outputs.total} projects in the workspace {self.name!r}. Only the first 10 projects are displayed in the printed table below:
            """
            ).lstrip()
            tips = dedent(
                """
            To access a specific project, use \033[1mworkspace\033[0m.project(Project ID)"""
            ).lstrip()
            link = dedent(
                f"""
                For quick access to the list of projects in the Vectice web app, visit:
                {self._client.auth.api_base_url}/browse/workspace/{self.id}"""
            ).lstrip()

            temp_print(description)
            temp_print(table=rich_table)
            temp_print(tips)
            temp_print(link)

        return project_list

    def create_project(
        self, name: str, description: str | None = None, template: str | None = None, copy_project_id: str | None = None
    ) -> Project:
        """Creates a project.

        Parameters:
            name: The project's name.
            description: The project's description.
            template: Creates the project from a spectific template. Available templates options : `tutorial`, `quickstart`, `default` (crisp-dm).
            copy_project_id: Creates the project from another project using the id `PRJ-XXX`. If template parameter is passed, it will take precedent over the project id.

        Returns:
            The newly created project.
        """
        if template is not None and copy_project_id is not None:
            _logger.warning(
                "Both template and copy_project_id parameters have been passed, template will take precedent over the project id."
            )

        item = self._client.create_project(
            self.id,
            {"name": name, "description": description, "template": template, "copy_project_id": copy_project_id},
        )
        logging_output = dedent(
            f"""
                Project {item.name!r} successfully created.

                For quick access to the Project in the Vectice web app, visit:
                {self._client.auth.api_base_url}/browse/project/{item.id}"""
        ).lstrip()
        _logger.info(logging_output)
        project_object = Project(item.id, self, item.name, item.description)
        return project_object

    @property
    def connection(self) -> Connection:
        """The Connection to which this workspace belongs.

        Returns:
            The Connection to which this workspace belongs.
        """
        return self._connection

    def list_templates(self, display_print: bool = True) -> list[ProjectTemplate]:
        """Retrieves a list of project templates which belong to the workspace. It will also print the first 10 project templates as a tabular form.

        Parameters:
            display_print: If set to True, it will print the first 10 project templates in a tabular form.

        Returns:
            A list of `ProjectTemplate` instances.
        """
        project_template_outputs = self._client.list_templates()
        project_template_list = [
            ProjectTemplate(project.name, project.description) for project in project_template_outputs
        ]

        if display_print:
            rich_table = Table(expand=True, show_edge=False)

            rich_table.add_column("Name", justify="left", no_wrap=True, max_width=15)
            rich_table.add_column("Description", justify="left", no_wrap=True, max_width=50)

            for project in project_template_outputs[:DEFAULT_PRINT_SIZE]:
                rich_table.add_row(project.name, format_description(project.description))

            description = dedent(
                f"""
            There are {len(project_template_outputs)} project templates that you can use to create a project. Only the first 10 project templates are displayed in the printed table below:
            """
            ).lstrip()

            temp_print(description)
            temp_print(table=rich_table)

        return project_template_list



from __future__ import annotations

import inspect
import logging
import re
import sys
import threading
from inspect import FrameInfo
from types import FrameType
from typing import Any, Dict, Set

_logger = logging.getLogger(__name__)


class CodeTraceService:
    """A service that automatically captures source code and variables of functions called within its context.

    The purpose of this service is to trace the execution of Python code, capturing the source code of functions
    as they are called, and to collect variables that are defined or modified during the execution of the traced code.

    This service is designed to be used as a context manager, allowing you to easily start and stop tracing
    without needing to manually manage the tracing state.

    See: https://docs.python.org/3/library/trace.html
    """

    def __init__(self, warn_on_trace: bool = True):
        self.warn_on_trace = warn_on_trace

        self._code = set()
        self._captured_functions = set()

        self._initial_globals = {}
        self._initial_locals = {}
        self._captured_variables = {}

        self._original_trace = None
        self._thread_id = None
        self._vectice_frame_index = 2

    @property
    def code(self) -> Set[str]:
        """Get all captured source code."""
        return self._code

    @property
    def captured_variables(self) -> Dict[str, Any]:
        """Get all captured variables."""
        return self._captured_variables

    def __enter__(self):
        """Enter the context manager and start tracing.

        Important notes:
        - This method sets the global trace function to capture function calls and their source code.
        - It also captures the initial global and local variables in the current frame.
        - If another trace function is already active, it will warn the user unless `warn_on_trace` is set to False.
        - The tracing is limited to the current thread to avoid interference with other threads or debugggers.

        Technical details:
        - Uses `sys.settrace()` to set a custom trace function that captures function calls.
        - Uses `sys._getframe()` to access the current frame and capture initial global and local variables.
        - Most importantly, the depth of the frame is set to 1 to capture the caller's frame, which is where the context manager is invoked.
        """
        # Flush any previously captured data, allows re-use of the service
        self.flush()
        # Check if tracing is already active
        current_trace = sys.gettrace()
        if current_trace is not None:
            # This warning will help for debugging purposes, but can be disabled if needed.
            if self.warn_on_trace:
                _logger.warning(
                    "Another trace function is already active. This may interfere with Vectice Trace.",
                )
            self._original_trace = current_trace

        try:
            stack = inspect.stack()
            vectice_frame: FrameType = self._find_vectice_trace_frame(stack)

            # Inital globals and locals, we use the difference between the initial and final globals/locals to capture changes.
            self._initial_globals = dict(vectice_frame.f_globals)
            self._initial_locals = dict(vectice_frame.f_locals)
            # We need to find the module name to get the source code of the module.
            module_name = vectice_frame.f_globals.get("__name__", "__main__")
            self._code.add(self._get_module_source_code(module_name))
            # Store the thread ID to ensure we only trace this thread
            self._thread_id = threading.get_ident()
            sys.settrace(self._safe_trace_calls)
        except Exception as e:
            # Restore original trace on error
            if self._original_trace:
                sys.settrace(self._original_trace)
            _logger.warning(f"Error during Vectice trace setup: {e}", exc_info=True)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # pyright: ignore[reportMissingParameterType]
        """Exit the context manager and stop tracing.

        Important notes:
        - This method restores the original trace function and captures the final global and local variables.
        - It ensures that the tracing state is cleaned up properly, even if an error occurs during tracing.
        - The final global and local variables are captured to allow inspection of the state after the traced code has executed.

        Technical details:
        - Uses `sys.settrace()` to restore the original trace function.
        - Uses `sys._getframe()` to access the caller's frame and capture the final global and local variables.
        - The depth of the frame is set to 2 to capture the caller's frame, which is where the context manager is invoked.
        - We clear local and global variables to prevent memory leaks.
        - Finally, it clears references to prevent memory leaks.
        - If an error occurs during cleanup, it logs a warning but does not prevent the cleanup from completing.
        """
        try:
            # Always restore the original trace function
            sys.settrace(self._original_trace)

            # Capture the final global and local variables
            vectice_frame: FrameType = sys._getframe(self._vectice_frame_index)  # pyright: ignore[reportPrivateUsage]

            final_globals = dict(vectice_frame.f_globals)
            self._capture_variables(final_globals)
            del final_globals

            final_locals = dict(vectice_frame.f_locals)
            self._capture_variables(final_locals)
            del final_locals

        except Exception as e:
            _logger.warning(f"Error during Vectice trace cleanup: {e}")

        finally:
            # Always clear references to prevent memory leaks
            self._original_trace = None
            self._thread_id = None

        return False

    #### Utility Methods ####

    def _find_vectice_trace_frame(self, frames: list[FrameInfo]) -> FrameType:
        """Find the frame where the Vectice trace context manager was invoked.

        The frames are added to the stack in reverse order, so the frame where the context manager is invoked
        is actually the frame after all the __enter__ frames.

        We look for the the context manager frames with `__enter__`.
        stack = [__enter__, __enter__, with vectice.trace(), ...]

        The length of the __enter__ frames `len(stack)` will be the index we entered the context manager.
        """
        frame_index = len([frame for frame in frames if re.search("__enter__", str(frame.function))])
        # set the vectice frame index to be used for __exit__
        self._vectice_frame_index = frame_index
        return frames[frame_index].frame

    def _capture_variables(self, variables: Dict[str, Any]):
        for name, value in variables.items():
            # Shallow copy of initial globals and locals to check for changes
            all_intials = {**self._initial_globals, **self._initial_locals}

            is_valid_variable = (
                not name.startswith("__")
                and "@py_assert" not in name
                and (name not in all_intials or all_intials[name] is not value)
            )
            if is_valid_variable:
                self._captured_variables[name] = value

    def _get_module_source_code(self, module_name: str = "__main__") -> str:
        """Get the source code of the specified module (defaults to main).

        This gets the code where the script was run, which is usually the main module.
        """
        module = sys.modules.get(module_name)

        if module is None:
            _logger.warning(
                "Could not get the module. It may not be loaded or does not exist.",
            )
            return ""

        try:
            return inspect.getsource(module)
        except (OSError, TypeError):
            _logger.warning(
                f"Could not get source code for module {module_name}. It may be a built-in module or not available.",
            )

        return ""

    def _is_user_function(self, filename: str | None = None, func_name: str | None = None):
        """Determine if a function is user-created vs from a library.

        Checks site-packages and standard library locations.
        """
        if not filename or not filename.endswith(".py"):
            return False

        # Skip obvious non-user functions
        if func_name and (func_name.startswith("<") or func_name == "<module>"):
            return False

        # Normalize path separators for cross-platform compatibility
        normalized_filename = filename.replace("\\", "/")

        # If it's in site-packages, it's library code
        if "site-packages" in normalized_filename:
            return False

        # we do not want to capture vectice code
        if "src/vectice" in normalized_filename:
            return False

        # Check for standard library patterns (cross-platform)
        stdlib_patterns = [
            "/lib/python",  # Unix-like systems
            "/Lib/",  # Windows
            "\\Lib\\",  # Windows with backslashes
        ]

        for pattern in stdlib_patterns:
            if pattern in filename and any(
                f"/python{version}" in filename or f"\\python{version}" in filename
                for version in ["3.8", "3.9", "3.10", "3.11", "3.12", "3.13"]
            ):
                return False

        # Everything else is user code
        return True

    def _capture_function_source(self, frame: FrameType, func_name: str):
        # Get the function object from the frame
        func_locals = frame.f_locals
        func_globals = frame.f_globals

        # Try to find the function in globals or locals
        func = func_globals.get(func_name) or func_locals.get(func_name)

        if func and callable(func):
            func_signature = inspect.getsource(func)
            self._code.add(func_signature)
            self._captured_functions.add(func_name)

    def flush(self):
        """Flush the captured code and variables.

        This method is a no-op in this context, as the captured code and variables are stored in memory.
        It can be used to clear the captured data if needed.
        """
        self._code.clear()
        self._captured_variables.clear()
        self._captured_functions.clear()
        self._initial_globals.clear()
        self._initial_locals.clear()

    ##### Trace Methods #####

    def _safe_trace_calls(self, frame: FrameType, event: str, arg: Any):
        """Safe wrapper around the trace function.

        The key points:
        1. **Call original first** - `self._original_trace(frame, event, arg)`
        2. **Respect original's return** - If it returns `None`, stop tracing
        3. **Chain the return** - Return your tracer or fall back to original
        4. **Error handling** - Always fall back to original tracer on errors
        """
        try:
            # Call the original tracer first (if it exists)
            if self._original_trace:
                original_result = self._original_trace(frame, event, arg)
                # If original tracer returns None, it wants to stop tracing
                if original_result is None:
                    return None

            # Only trace the correct thread
            # TODO: This might be unreliable according to the documentation
            if threading.get_ident() != self._thread_id:
                return self._safe_trace_calls

            self._trace_calls(frame, event, arg)

            # We keep calling the safe trace function to continue tracing
            return self._safe_trace_calls

        except Exception as e:
            _logger.warning(f"Error in trace function: {e}")
            # Fall back to original tracer on error
            return self._original_trace

    def _trace_calls(self, frame: FrameType, event: str, arg: Any):
        """Trace function calls and capture their source.

        A function is called (or some other code block entered). The global trace function is called; arg is None;
        the return value specifies the local trace function.

        See: https://docs.python.org/3/library/sys.html#sys.settrace
        """
        func_name = frame.f_code.co_name
        filename = frame.f_code.co_filename

        # Skip built-in, site-packages and builtins
        is_user_func = self._is_user_function(filename, func_name)
        if not is_user_func:
            return self._safe_trace_calls

        if event == "call":
            # Skip functions we have seen already
            is_function_to_capture = func_name not in self._captured_functions

            if is_function_to_capture:
                self._capture_function_source(frame, func_name)
        elif event == "return":
            # We capture the local scope of returns
            self._capture_variables(frame.f_locals)

        return self._safe_trace_calls



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Sequence

from giskard.rag import AgentAnswer, KnowledgeBase, QATestset, evaluate, generate_testset  # type: ignore[reportMissingImports]

if TYPE_CHECKING:
    from giskard.rag import RAGReport  # type: ignore[reportMissingImports]
    from pandas import DataFrame

_logger = logging.Logger(__name__)


def giskard_rag_wrapper(
    knowledge_base: KnowledgeBase | DataFrame | None = None,
    chat_engine: Any | None = None,
    answer_function: Any | Sequence[AgentAnswer | str] | None = None,
    qa_testset: QATestset | str | None = None,
    testset_num_questions: int | None = None,
    metrics: list[Any] | None = None,
    agent_description: str | None = None,
) -> RAGReport:
    """Wrapper function to configure and evaluate a Retrieval-Augmented Generation (RAG) agent
    using Giskard's RAG module and Azure's Chat Engine.

    This function sets up the RAG agent by initializing necessary configurations, including
    extracting the index, setting up the knowledge base, and the QA testset. The function
    then evaluates the agent using pre-defined RAGAS metrics.

    Args:
        index_zip_path (str): The file path to the zipped index containing preprocessed data.
        index_extract_dir (str): The directory where the index will be extracted.
        knowledge_base_path (str): The path to the knowledge base for retrieval purposes.
        qa_testset_path (str): The file path to the QA test set used for evaluating the RAG agent.
        custom_answer_fn (Callable | None, optional): Custom function to override the default
            answer function. Defaults to None.

    Returns:
        RAGReport: A report containing the evaluation results of the RAG agent,
        including relevant metrics and performance insights.

    Example:
        report = giskard_rag_wrapper(
            index_zip_path="/path/to/index.zip",
            index_extract_dir="/path/to/extracted/index",
            knowledge_base_path="/path/to/knowledge_base",
            qa_testset_path="/path/to/qa_testset.json",
        )
    """
    from pandas import DataFrame

    if knowledge_base is None and qa_testset is None:
        raise ValueError("knowledge_base or qa_testset must be provided")
    if isinstance(knowledge_base, DataFrame):
        knowledge_base = KnowledgeBase(knowledge_base)

    if isinstance(qa_testset, str):
        giskard_testset = QATestset.load(qa_testset)
    elif isinstance(qa_testset, QATestset):
        giskard_testset = qa_testset
    elif qa_testset is None and knowledge_base is not None:
        giskard_testset = generate_testset(knowledge_base, testset_num_questions, agent_description=agent_description)
    else:
        raise ValueError("Unable to generate qa_testset, please check knowledge_base or qa_testset provided.")

    if answer_function is None and chat_engine is None:
        raise ValueError("The answer function or chat engine must be provided")

    if answer_function is None:
        _logger.info("If the answer_function is not provided. A generic answer function will be used.")
        set_chat_engine(chat_engine)
        answer_function = get_answer_fn

    return evaluate(answer_function, testset=giskard_testset, knowledge_base=knowledge_base, metrics=metrics)


global_chat_engine = None


def set_chat_engine(chat_engine: Any):
    global global_chat_engine
    # Initialize your chat engine here
    global_chat_engine = chat_engine


def answer_fn(question: str, history: Any = None) -> Any:
    try:
        from llama_index.core.base.llms.types import ChatMessage, MessageRole  # type: ignore[reportMissingImports]
    except ImportError:
        raise ImportError("Please pip install llama_index to use the generic answer functionality.")

    global global_chat_engine
    if global_chat_engine is None:
        raise ValueError("Chat engine not initialized. Call initialize_chat_engine() first.")

    if history:
        answer = global_chat_engine.chat(
            question,
            chat_history=[
                ChatMessage(
                    role=MessageRole.USER if msg["role"] == "user" else MessageRole.ASSISTANT,
                    content=msg["content"],
                )
                for msg in history
            ],
        )
    else:
        answer = global_chat_engine.chat(question, chat_history=[])
    return answer


def get_answer_fn(question: str, history: Any = None) -> AgentAnswer:
    """A function representing your RAG agent."""
    # Get the answer and the documents
    agent_output = answer_fn(question, history)

    # Following llama_index syntax, you can get the answer and the retrieved documents
    answer = agent_output.response
    documents = agent_output.source_nodes
    document_nodes = [node.text for node in documents]

    # Instead of returning a simple string, we return the AgentAnswer object which
    # allows us to specify the retrieved context which is used by RAGAS metrics
    return AgentAnswer(message=answer, documents=document_nodes)



from __future__ import annotations

import csv
import functools
import logging
from datetime import datetime
from typing import Union

from typing_extensions import get_args

from vectice.api.http_error_handlers import VecticeException
from vectice.models.attachment import TAttachment, TFormattedAttachment
from vectice.models.property import Property
from vectice.models.representation.dataset_representation import DatasetRepresentation
from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
from vectice.models.resource.base import Resource
from vectice.models.resource.metadata.base import DatasetSourceType, DatasetSourceUsage, DatasetType, MetadataSettings
from vectice.models.resource.snowflake_resource import SnowflakeResource
from vectice.models.resource.spark_table_resource import SparkTableResource
from vectice.utils.common_utils import format_attachments, format_properties

_logger = logging.getLogger(__name__)

TBaseDerivedFrom = Union[str, DatasetRepresentation, DatasetVersionRepresentation]


class Dataset:
    def __init__(
        self,
        type: DatasetType,
        name: str | None = None,
        resource: Resource | None = None,
        training_resource: Resource | None = None,
        testing_resource: Resource | None = None,
        validation_resource: Resource | None = None,
        derived_from: list[TBaseDerivedFrom | Dataset] | TBaseDerivedFrom | Dataset | None = None,
        properties: dict[str, str | int] | list[Property] | Property | None = None,
        attachments: TAttachment | None = None,
    ):
        self._type = type
        self._name = name or f"dataset {datetime.now()}"
        self._resource = resource
        self._training_resource = training_resource
        self._testing_resource = testing_resource
        self._validation_resource = validation_resource
        self._derived_from = get_derived_from(derived_from)
        self._latest_version_id: str | None = None
        self._properties = format_properties(properties) if properties else None
        self._attachments = format_attachments(attachments) if attachments else None

        if self._type is DatasetType.MODELING:
            if self._training_resource is None or self._testing_resource is None:
                raise ValueError("You cannot create a modeling dataset without both training and testing sets")

            self._training_resource.usage = DatasetSourceUsage.TRAINING
            self._testing_resource.usage = DatasetSourceUsage.TESTING
            if self._validation_resource:
                self._validation_resource.usage = DatasetSourceUsage.VALIDATION

        self._has_snowflake_resource = (
            isinstance(self._resource, SnowflakeResource)
            or isinstance(self._training_resource, SnowflakeResource)
            or isinstance(self._testing_resource, SnowflakeResource)
            or isinstance(self._validation_resource, SnowflakeResource)
        )
        self._has_spark_resource = (
            isinstance(self._resource, SparkTableResource)
            or isinstance(self._training_resource, SparkTableResource)
            or isinstance(self._testing_resource, SparkTableResource)
            or isinstance(self._validation_resource, SparkTableResource)
        )

        metadata_dict, is_updating_columns = self._transform_resource_to_metadata_dict()
        self._is_updating_columns = is_updating_columns
        self._metadata_as_dict = metadata_dict

    @staticmethod
    def origin(
        resource: Resource,
        name: str | None = None,
        properties: dict[str, str | int] | list[Property] | Property | None = None,
        attachments: str | list[str] | None = None,
    ) -> Dataset:
        """Create an origin dataset.

        Examples:
            ```python
            from vectice import Dataset, FileResource

            dataset = Dataset.origin(
                name="my origin dataset",
                resource=FileResource(paths="origin_dataset.csv"),
            )
            ```

        Parameters:
            resource: The resource for the origin dataset.
            name: The name of the dataset.
            properties: A dict, for example `{"folds": 32}`.
            attachments: The file paths that will be attached to the iteration along with the dataset.
        """
        return Dataset(
            type=DatasetType.ORIGIN,
            name=name,
            resource=resource,
            properties=properties,
            attachments=attachments,
        )

    @staticmethod
    def clean(
        resource: Resource,
        name: str | None = None,
        derived_from: list[TBaseDerivedFrom | Dataset] | TBaseDerivedFrom | Dataset | None = None,
        properties: dict[str, str | int] | list[Property] | Property | None = None,
        attachments: str | list[str] | None = None,
    ) -> Dataset:
        """Create a clean dataset.

        Examples:
            ```python
            from vectice import Dataset, FileResource

            dataset = Dataset.clean(
                name="my clean dataset",
                resource=FileResource(paths="clean_dataset.csv"),
            )
            ```

        Parameters:
            resource: The resource for the clean dataset.
            name: The name of the dataset.
            derived_from: List of dataset versions or model versions (or their ids) from which this dataset is derived.
            properties: A dict, for example `{"folds": 32}`.
            attachments: The file paths that will be attached to the iteration along with the dataset.
        """
        return Dataset(
            type=DatasetType.CLEAN,
            name=name,
            resource=resource,
            derived_from=derived_from,
            properties=properties,
            attachments=attachments,
        )

    @staticmethod
    def modeling(
        training_resource: Resource,
        testing_resource: Resource,
        validation_resource: Resource | None = None,
        name: str | None = None,
        properties: dict[str, str | int] | list[Property] | Property | None = None,
        attachments: str | list[str] | None = None,
        derived_from: list[TBaseDerivedFrom | Dataset] | TBaseDerivedFrom | Dataset | None = None,
    ) -> Dataset:
        """Create a modeling dataset.

        Examples:
            ```python
            from vectice import Dataset, FileResource

            dataset = Dataset.modeling(
                name="my modeling dataset",
                training_resource=FileResource(paths="training_dataset.csv"),
                testing_resource=FileResource(paths="testing_dataset.csv"),
                validation_resource=FileResource(paths="validation_dataset.csv"),
            )
            ```

        Parameters:
            training_resource: The resource for the training set (for modeling datasets).
            testing_resource: The resource for the testing set (for modeling datasets).
            validation_resource: The resource for the validation set (optional, for modeling datasets).
            name: The name of the dataset.
            properties: A dict, for example `{"folds": 32}`.
            attachments: The file paths that will be attached to the iteration along with the dataset.
            derived_from: List of dataset versions or model versions (or their ids) from which this dataset is derived.
        """
        return Dataset(
            type=DatasetType.MODELING,
            name=name,
            training_resource=training_resource,
            testing_resource=testing_resource,
            validation_resource=validation_resource,
            properties=properties,
            attachments=attachments,
            derived_from=derived_from,
        )

    def __repr__(self):
        return f"Dataset(name={self.name!r}, type={self.type!r})"

    @property
    def type(self) -> DatasetType:
        """The dataset's type.

        Returns:
            The dataset's type.
        """
        return self._type

    @property
    def resource(self) -> Resource | tuple[Resource, Resource, Resource | None]:
        """The dataset's resource.

        Returns:
            The dataset's resource.
        """
        if self._type is DatasetType.MODELING:
            return self._training_resource, self._testing_resource, self._validation_resource  # type: ignore[return-value]
        return self._resource  # type: ignore[return-value]

    @property
    def name(self) -> str:
        """The dataset's name.

        Returns:
            The dataset's name.
        """
        return self._name

    @name.setter
    def name(self, name: str):
        """Set the dataset's name.

        Parameters:
            name: The name of the dataset.
        """
        self._name = name

    @property
    def derived_from(self) -> list[str]:
        """The datasets versions or models versions from which this dataset is derived.

        Returns:
            The datasets versions or models versions from which this dataset is derived.
        """
        return self._derived_from

    @property
    def latest_version_id(self) -> str | None:
        """The id of the latest version of this dataset.

        Returns:
            The id of the latest version of this dataset.
        """
        return self._latest_version_id

    @latest_version_id.setter
    def latest_version_id(self, value: str) -> None:
        """Set the id of the latest version of this dataset.

        Parameters:
            value: The id of the latest version of this dataset.
        """
        self._latest_version_id = value

    @property
    def properties(self) -> list[Property] | None:
        """The dataset's properties.

        Returns:
            The dataset's properties.
        """
        return self._properties

    @properties.setter
    def properties(self, properties: dict[str, str | int] | list[Property] | Property | None):
        """Set the dataset's properties.

        Parameters:
            properties: The properties of the dataset.
        """
        _logger.warning("To save your updated dataset properties, you must reassign your dataset to an iteration.")
        self._properties = format_properties(properties) if properties else None

    @property
    def attachments(self) -> list[TFormattedAttachment] | None:
        """The file attachment objects or paths associated with the dataset.

        Returns:
        `List[Union[Table, str]] | None`
            A list of attachment objects, where each attachment is one of:

        - `Table`
            A formatted table object containing structured data
        - `str`
            A string representation of an attachment (e.g., file path)
        - `None`
            If no attachments are present
        """
        return self._attachments

    @attachments.setter
    def attachments(self, attachments: TAttachment):
        """Attach a file or files to the dataset.

        Parameters:
            attachments: The filename or filenames of the file or set of files to attach to the dataset.
        """
        self._attachments = format_attachments(attachments)

    def _transform_resource_to_metadata_dict(self):
        from vectice.api.client import Client

        resources = self.resource if isinstance(self.resource, tuple) else [self.resource]
        metadata_asdict = []
        is_updating_columns = False

        org_config = Client.get_instance().org_config
        settings = MetadataSettings(
            minimum_rows_for_statistics=org_config.configuration.df_statistics_row_threshold,
            sample_rows_for_statistics=org_config.configuration.df_statistics_row_sample,
            maximum_columns_for_statistics=org_config.configuration.df_statistics_colmns_threshold,
        )

        for resource in resources:
            if resource is not None and resource.metadata:
                resource.metadata.set_settings(settings)

                metadata, columns_description = resource.metadata, resource.columns_description
                is_updating_columns = is_updating_columns is False and columns_description is not None
                columns = columns_description or {}
                metadata_dict, type = metadata.asdict(), metadata.type
                if type == DatasetSourceType.DB:
                    for db in metadata_dict["dbs"]:
                        db["columns"] = self._combine_source_column_description(db, columns)
                else:
                    for file in metadata_dict["files"]:
                        file["columns"] = self._combine_source_column_description(file, columns)

                metadata_asdict.append(metadata_dict)
        return metadata_asdict, is_updating_columns

    def _combine_source_column_description(self, source: dict, columns_description: dict[str, str] | str):
        def _transform_path_to_dictionary(path: str):
            def _reduce_csv(acc: dict[str, str], curr: list[str]):
                if len(curr) > 1:
                    return {**acc, curr[0]: curr[1]}
                return acc

            with open(path, mode="r", newline="") as csv_file:
                return functools.reduce(_reduce_csv, csv.reader(csv_file), {})

        if isinstance(columns_description, str):
            columns_description = _transform_path_to_dictionary(columns_description)

        def _map_columns(col: dict):
            col_name = col["name"]
            if col_name in columns_description:
                column_desc = columns_description[col_name]
                return {**col, "description": column_desc}
            return col

        return list(map(_map_columns, source["columns"]))


TDerivedFrom = Union[TBaseDerivedFrom, Dataset]


def get_derived_from(
    derived_from: list[TDerivedFrom] | TDerivedFrom | None,
) -> list[str]:

    from vectice.models.model import Model
    from vectice.models.representation.model_representation import ModelRepresentation
    from vectice.models.representation.model_version_representation import ModelVersionRepresentation

    derived_from_ids: list[str] = []
    formatted_df = _format_derived_from(derived_from)
    for df in formatted_df or []:
        if isinstance(df, Dataset):
            if df.latest_version_id is None:
                raise ValueError(
                    f"Dataset {df.name!r} does not have a version id. "
                    "Was it registered in Vectice (logged to an iteration)?"
                )
            derived_from_ids.append(df.latest_version_id)
        elif isinstance(df, Model):
            if df.latest_version_id is None:
                raise ValueError(
                    f"Model {df.name!r} does not have a version id. "
                    "Was it registered in Vectice (logged to an iteration)?"
                )
            derived_from_ids.append(df.latest_version_id)
        elif isinstance(df, DatasetRepresentation) or isinstance(df, ModelRepresentation):
            last_version = df._last_version  # pyright: ignore[reportPrivateUsage]
            if last_version is None:
                raise VecticeException("Unable to log asset, no versions exist.")
            derived_from_ids.append(last_version.id)
        elif isinstance(df, DatasetVersionRepresentation) or isinstance(df, ModelVersionRepresentation):
            derived_from_ids.append(df.id)
        elif isinstance(df, str):  # pyright: ignore[reportUnnecessaryIsInstance]
            if df.startswith("DTV-") or df.startswith("MDV-"):
                derived_from_ids.append(df)
            else:
                raise VecticeException(f"Please check the derived_from id {df}.")
    return derived_from_ids


def _format_derived_from(derived_from: list[TDerivedFrom] | TDerivedFrom | None) -> list[TDerivedFrom]:
    if derived_from is None:
        return []

    formatted_derived_from = derived_from if isinstance(derived_from, list) else [derived_from]

    for df in formatted_derived_from:
        if not isinstance(df, get_args(TDerivedFrom)):
            raise ValueError("Invalid derived_from parameter")

    return formatted_derived_from
